(*
   Title  : Bag.thy
   Author : M. Bortin
*)

header {* Implementation of finite multisets w.r.t. Bag-Spec *}



theory Bag
imports "../../Extensions/AWE_HOL" 
begin



consts
 count :: "'a \<Rightarrow> 'a list \<Rightarrow> nat"
primrec
"count a [] = 0"
"count a (x#xs) = (if x = a then Suc(count a xs) else count a xs)"


lemma count0 : " (count x xs = 0) = (x \<notin> set xs) " 
by(induct xs, simp_all)

lemma count_append[rule_format] :
" \<forall>ys. count x (xs @ ys) = count x xs + count x ys"
by(induct_tac xs, simp_all)


lemma remove1_fix[rule_format] : "x \<notin> set xs \<longrightarrow> remove1 x xs = xs"
by(induct_tac xs, simp_all)

lemma remove1_count1 : "a \<noteq> b \<Longrightarrow> count a (remove1 b xs) = count a xs"
by(induct_tac xs, simp_all)

lemma remove1_count2[rule_format] : "a \<in> set xs \<longrightarrow> count a (remove1 a xs) = count a xs - 1"
by(induct_tac xs, simp_all)

section {* Equivalence relation on lists induced by @{term "count"} *}

text 
{* The equivalence relation is the intersection of kernels of 
   the family of functions @{text "count a"} where @{text "a"} ranges
   over the universe of some fixed arbitrary type @{text "\<tau>"}. 
*}      

constdefs
 eqR :: "'a list \<times> 'a list \<Rightarrow> bool"
"eqR \<equiv> (\<lambda>(xs, ys). (\<forall>a. count a xs = count a ys))"


sigmorph bag : QuotientType_Parameters \<longrightarrow> Bag
type_map: ["'a T" \<mapsto> "'a list"]

thymorph_goals bag

lemma eqR_sym:
"a \<simeq> b \<Longrightarrow> b \<simeq> a"
by(simp add: eqR_def mem_def)

lemma eqR_refl:
"a \<simeq> a"
by(simp add: eqR_def mem_def)

lemma eqR_trans:
"\<lbrakk>a \<simeq> b; b \<simeq> c\<rbrakk> \<Longrightarrow> a \<simeq> c"
by(simp add: eqR_def mem_def)

thymorph bag by_sigmorph bag


section {* Some basic properties of @{term "eqR"} *}

lemma eqR_eq : "(s \<simeq> t) = (\<forall>x. count x s = count x t)"
by(simp add: eqR_def mem_def)

lemma eqR_mem : " s \<simeq> t \<Longrightarrow> x \<in> set s \<Longrightarrow> x \<in> set t"
apply(rule ccontr)
apply(subst (asm) count0[THEN sym])
apply(subst (asm) eqR_eq)
apply(drule_tac x=x in spec, simp)
apply(subst (asm) count0)+
by clarify

lemma eqR_Nil : "s \<simeq> Nil \<Longrightarrow> s = Nil"
apply(subst (asm) eqR_eq)
apply(rule ccontr)
apply(subst (asm) hd_Cons_tl[THEN sym, of s], assumption)
apply(drule_tac x="hd s" in spec)
apply(subst (asm) count.simps)+
by simp



lemma eqR_remove1 :
"a \<in> set xs \<Longrightarrow> xs \<simeq> a # (remove1 a xs)"
apply(subst eqR_eq)
apply(rule allI)
apply(case_tac "x = a", clarsimp)
apply(subst remove1_count2, assumption)
apply(case_tac "count a xs = 0")
apply(subst (asm) count0)
apply clarify
apply simp
apply simp
apply(subst remove1_count1, assumption)
by(rule refl)


section {* Getting the associated quotient type *}

instantiate_theory QuotientType by_thymorph bag
renames: [("QuotientType.TQ" \<mapsto> Bag)]


translate_thm "QuotientType.ext1" along bag'
translate_thm "QuotientType.ext2" along bag'
translate_thm "QuotientType.class_of_eq" along bag'
translate_thm "QuotientType.class_of_repr" along bag'




section {* Deriving fold for bags *}


instantiate_theory Fold_Parameters by_thymorph bag'
type_map: ["('u, 'a) \<Sigma>" \<mapsto> "unit + 'u \<times> 'a"]



defs
 TC_def : "TC xs \<equiv> case xs of
                    Nil \<Rightarrow> Inl ()
                  | x#xs \<Rightarrow> Inr (x, xs)"

 Sigma_def : "\<Sigma> f q \<equiv> (case q of 
                        Inl u \<Rightarrow> Inl u
                      | Inr (a, v) \<Rightarrow> Inr (a, f v))"

 Sigma'_def : "\<Sigma>' X \<equiv> {(Inl (), Inl ())} \<union> {(Inr (a, u), Inr (a, v)) | a u v. (u, v) \<in> X}"


 Transform_def : "Transform s t \<equiv> (case t of 
                                    Inl u \<Rightarrow> Inl u
                                  | Inr (x, xs) \<Rightarrow> (case s of
                                                     Inl u \<Rightarrow> Inr (x, xs)
                                                   | Inr (y, ys) \<Rightarrow> Inr (y, remove1 y (x#xs))))"

 FoldCond_def : "FoldCond \<equiv> UNIV" 

 TransformCond_def : "TransformCond \<equiv> {A. \<forall>a b c. A (Inr (a, A (Inr (b, c)))) = A (Inr (b, A (Inr (a, c)))) }"

 Fold_def : "Fold A xs \<equiv> foldr (\<lambda>a b. A (Inr (a, b))) xs (A (Inl ()))"  


lemma TransformCond:
"TransformCond \<subseteq> FoldCond"
by(simp add: FoldCond_def)


lemma Fold_eq:
"A \<in> FoldCond \<Longrightarrow> Fold A = A \<circ> \<Sigma> (Fold A) \<circ> TC"
apply(rule ext, simp)
apply(case_tac x, clarify)
apply(simp add: TC_def)
apply(simp add: Sigma_def)
apply(simp add: Fold_def)
apply clarify
apply(simp add: TC_def)
apply(simp add: Sigma_def)
apply(subst Fold_def)
apply simp
apply(fold Fold_def)
by(rule refl)


lemma TC_reduct:
"TC -` \<Sigma>'' P \<subseteq> P \<Longrightarrow> UNIV \<subseteq> P"
apply clarsimp
apply(subst (asm) subset_iff)
apply(induct_tac x)
apply(drule spec, erule mp)
apply(simp add: TC_def)
apply(simp add: Sigma''_def monoP_def)
apply(simp add: Sigma'_def)
apply(drule spec, erule mp)
apply(simp add: TC_def)
apply(simp add: Sigma''_def monoP_def)
apply(simp add: Sigma'_def)
by(simp add: monoE_def)

lemma Transform:
"s \<simeq> t \<Longrightarrow> TC s \<simeq>\<^sub>\<Sigma> Transform (TC s) (TC t)"
apply(case_tac s, clarify)
apply(case_tac t, clarify)
apply(simp add: TC_def Transform_def eqR_ext_def)
apply(simp add: Sigma'_def)
apply clarify
apply(drule eqR_sym)
apply(drule_tac x=a in eqR_mem, simp)
apply simp
apply(case_tac t, clarify)
apply(drule_tac x=a in eqR_mem, simp)
apply simp
apply clarify
apply (simp add: TC_def Transform_def)
apply(rule conjI)
apply clarify
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply(subst eqR_eq)
apply(rule allI)
apply(subst (asm) eqR_eq)
apply(drule_tac x=x in spec)
apply(simp split: split_if_asm)
apply(rule impI)
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply(subst eqR_eq)
apply(rule allI)
apply(case_tac "x = aa", clarify)
apply(simp)
apply(subst remove1_count1, simp)
apply(subst (asm) eqR_eq)
apply(drule_tac x=aa in spec, simp)
apply simp
apply(case_tac "x = a", clarify)
apply(frule_tac x=a in eqR_mem, simp)
apply simp
apply(subst remove1_count2, assumption)
apply(subst (asm) eqR_eq)
apply(drule_tac x=a in spec, simp)
apply(subst remove1_count1, simp)
apply(subst (asm) eqR_eq)
apply(drule_tac x=x in spec, simp)
done


lemma Sigma_division :
"\<Sigma>'' (eqR \\ ker f) \<subseteq> eqR_ext \\ ker (\<Sigma> f)"
apply(simp add: division_def ker_def)
apply clarify
apply(case_tac x, clarify)
apply(case_tac b, clarify)
apply(simp add: eqR_ext_def)
apply clarify
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply clarify
apply(case_tac b, clarify)
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply clarify
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply clarify
apply(simp add: Sigma_def)
apply(simp add: Sigma''_def monoP_def)
apply(simp add: Sigma'_def)
apply(simp add: monoE_def)
done



lemma TransformCond_subst :
"A \<in> TransformCond \<Longrightarrow> A (Inr (a, A (Inr (b, c)))) = A (Inr (b, A (Inr (a, c))))"
by(simp add: TransformCond_def)


lemma Fold_Transform_aux[rule_format] :
"A \<in> TransformCond \<Longrightarrow> 
x \<in> set xs \<longrightarrow> Fold A (x # remove1 x xs) = Fold A xs"
apply(induct_tac xs, simp)
apply clarsimp
apply(simp (no_asm) add: Fold_def)
apply(subst TransformCond_subst[where A=A], assumption)
apply(subst foldr.simps(2)[where f="(\<lambda>a b. A (Inr (a, b)))" and a="A (Inl ())", THEN sym])
apply(fold Fold_def)
apply(erule ssubst)
by(rule refl)


lemma Fold_Transform:
"\<lbrakk>A \<in> TransformCond; s \<simeq> t\<rbrakk>
\<Longrightarrow> A (\<Sigma> (Fold A) (Transform (TC s) (TC t))) = Fold A t"
apply(case_tac s, clarify)
apply(case_tac t, clarify)
apply(simp add: TC_def)
apply(simp add: Transform_def)
apply(simp add: Sigma_def Fold_def)
apply clarify
apply(drule eqR_sym)
apply(drule_tac x=a in eqR_mem, simp)
apply simp
apply(case_tac t, clarify)
apply(drule_tac x=a in eqR_mem, simp)
apply simp
apply clarify
apply(simp add: TC_def)
apply(simp add: Transform_def)
apply(simp add: Sigma_def)
apply(rule conjI, clarsimp)
apply(subst Fold_def, simp)
apply(fold Fold_def)
apply(rule refl)
apply(rule impI)
apply(subst Fold_def)
apply simp
apply(subst TransformCond_subst[where A=A], assumption)
apply(subst foldr.simps(2)[where f="(\<lambda>a b. A (Inr (a, b)))" and a="A (Inl ())", THEN sym])
apply(fold Fold_def)
apply(drule_tac x=a in eqR_mem, simp)
apply simp
apply(subst Fold_Transform_aux, assumption+)
apply(subst Fold_def, simp)
apply(fold Fold_def)
by(rule refl)


thymorph fold_bag by_sigmorph bag''

instantiate_theory Fold by_thymorph fold_bag
renames: [("Fold.FoldQ" \<mapsto> FoldBag),
          ("Fold.FoldQ_def" \<mapsto> FoldBag_def)]

translate_thm "Fold.FoldQ_eq" as FoldBag_eq along fold_bag'



section {* Proving an instance of Bag_Spec *}


thymorph i : Nat \<longrightarrow> Bag (* an inclusion *)

instantiate_theory Bag_Spec by_thymorph i
type_map: ["'a bag" \<mapsto> "'a Bag"] 
renames: [("FiniteSet_Spec.Empty" \<mapsto> Empty_Bag mixfix: ("{#}"))] 


defs
 Empty_Bag_def   : "{#} \<equiv> <Nil>"

 Insert_def       : "Insert a \<equiv> (class_of \<circ> Cons a)\<^sup>#"

 Count_def        : "Count a \<equiv> (count a)\<^sup>#"

 Remove_def       :  "(M -- a) \<equiv>  (class_of \<circ> remove1 a)\<^sup># M"

 foldBag_def      : "foldBag f e \<equiv> FoldBag (\<lambda>x. case x of Inl u \<Rightarrow> e | Inr (a, b) \<Rightarrow> f a b) "


subsection {* Proving congruences *}

lemma count_congr :
"congruence (count a)"
apply(simp add: congruence_def, clarify)
apply(subst (asm) eqR_eq)
by simp

lemma Count_eq :
"(a <: <xs>) = count a xs" 
apply(unfold Count_def)
apply(subst ext1, rule count_congr)
by(rule refl)

lemma Cons_congr :
"congruence (class_of \<circ> Cons a)"
apply(simp add: congruence_def, clarify)
apply(rule class_of_eq[THEN iffD1])
apply(subst eqR_eq)
apply clarsimp
apply(subst (asm) eqR_eq)
by simp

lemma Insert_eq :
"(a ## <xs>) = <(a # xs)>"
apply(unfold Insert_def)
apply(subst ext2, rule Cons_congr)
by(rule refl)

lemma remove1_congr :
"congruence(class_of \<circ> remove1 a)"
apply(simp add: congruence_def, clarify)
apply(rule class_of_eq[THEN iffD1])
apply(subst eqR_eq)
apply(rule allI)
apply(case_tac "x = a", clarsimp)
apply(case_tac "a \<in> set s")
apply(frule eqR_mem, assumption)
apply(subst remove1_count2, assumption)+
apply(subst (asm) eqR_eq)
apply simp
apply(subgoal_tac "a \<notin> set t")
apply(subst remove1_fix, assumption)+
apply(subst (asm) eqR_eq, simp)
apply(rule notI)
apply(drule eqR_sym)
apply(drule eqR_mem, assumption, clarify)
apply(subst remove1_count1, assumption)+
apply(subst (asm) eqR_eq, simp)
done

lemma Remove_eq :
"(<xs> -- a) = <remove1 a xs>"
apply(unfold Remove_def)
apply(subst ext2, rule remove1_congr)
by(rule refl)


subsection {* Proving the theory morphism goals *}

 lemma Ext:
"(\<And>a. (a <: M) = (a <: T)) \<Longrightarrow> M = T"
apply(subst (asm) class_of_repr[THEN sym, of M])
apply(subst (asm) class_of_repr[THEN sym, of T])
apply(subst (asm) Count_eq)+
apply(drule eqR_eq[THEN iffD2, rule_format])
apply(subst (asm) class_of_eq)
apply(subst (asm) class_of_repr)+
by assumption


lemma Count_Empty:
"a <: {#} = 0"
apply(unfold Empty_Bag_def)
apply(subst Count_eq)
by simp

lemma Count_Insert_eq1:
"(a <: a ## M) = (a <: M) + 1"
apply(subst class_of_repr[THEN sym, of M])
apply(subst Insert_eq)
apply(subst Count_eq)
apply simp
apply(subst Count_eq[THEN sym])
apply(subst class_of_repr)
apply(rule refl)
done

lemma Count_Insert_eq2:
"a \<noteq> b \<Longrightarrow> (a <: b ## M) = (a <: M)"
apply(subst class_of_repr[THEN sym, of M])
apply(subst Insert_eq)
apply(subst Count_eq)
apply simp
apply(subst Count_eq[THEN sym])
apply(subst class_of_repr)
apply(rule refl)
done

lemma Count_Remove_eq1:
"(a <: M -- a) = (a <: M) - 1"
apply(subst class_of_repr[THEN sym, of M])
apply(subst Remove_eq)
apply(subst Count_eq)
apply simp
apply(case_tac "a \<in> set ^M")
apply(subst remove1_count2, assumption)
apply(subst Count_eq[THEN sym])
apply(subst class_of_repr)
apply simp
apply(subst remove1_fix, assumption)
apply(subst (asm) count0[THEN sym])
apply(subst Count_eq[THEN sym])
apply(subst class_of_repr)
apply(subst (asm) Count_eq[THEN sym])
apply(subst (asm) class_of_repr)
by simp


lemma Count_Remove_eq2:
"a \<noteq> b \<Longrightarrow> (a <: M -- b) = (a <: M)"
apply(subst class_of_repr[THEN sym, of M])
apply(subst Remove_eq)
apply(subst Count_eq)
apply(subst remove1_count1, assumption)
apply(subst Count_eq[THEN sym])
apply(subst class_of_repr)
by(rule refl)


lemma foldBag1:
"foldBag f e {#} = e"
apply(unfold foldBag_def)
apply(unfold FoldBag_def)
apply(unfold Empty_Bag_def)
apply(simp add: qtype_ext_def)
apply(subgoal_tac "^<[]> = []")
apply(erule ssubst)
apply(simp add: Fold_def)
apply(rule eqR_Nil)
apply(rule class_of_eq[THEN iffD2])
apply(subst class_of_repr)
by(rule refl)

lemma foldBag2:
"f \<in> leftCommuting \<Longrightarrow> foldBag f e (x ## M) = f x (foldBag f e M)"
apply(subst class_of_repr[THEN sym, of M])
apply(subst Insert_eq)
apply(subst foldBag_def)
apply(subst FoldBag_eq)
apply(simp add: TransformCond_def)
apply(simp add: leftCommuting_def)
apply(unfold Fold_def)
apply(subst foldr.simps)
apply(fold Fold_def)
apply(simp)
apply(subst FoldBag_eq[THEN sym])
apply(simp add: TransformCond_def)
apply(simp add: leftCommuting_def)
apply(fold foldBag_def)
apply(subst class_of_repr)
by(rule refl)


thymorph bag_spec by_sigmorph i'

translate_thm "Bag_Spec.Remove_induct" along bag_spec
translate_thm "Bag_Spec.MUnion_eq" along bag_spec
translate_thm "Bag_Spec.Map_le" along bag_spec
translate_thm "Bag_Spec.Map_inj" along bag_spec


end
