(*
   Title  : Bag_Spec.thy
   Author : M. Bortin
*)

header {* Specification and properties of finite multisets *}


theory Bag_Spec
imports Nat
begin

typedecl 'a bag


section {* Basic operations on @{typ "'a bag"} *}

consts
 Empty     :: "'a bag"                         ("{#}")
 Count     :: "'a \<Rightarrow> 'a bag \<Rightarrow> nat"            ("_ <: _" [50, 51] 50) 
 Insert    :: "'a \<Rightarrow> 'a bag \<Rightarrow> 'a bag"         ("_ ## _" [51, 52] 51) 
 Remove    :: "'a bag \<Rightarrow> 'a \<Rightarrow> 'a bag"         ("_ -- _" [53, 52] 52)


axioms
 Count_Empty        :   "a <: {#} = 0"
 Count_Insert_eq1   :   "a <: a ## M = (a <: M) + 1"
 Count_Insert_eq2   :   "a \<noteq> b \<Longrightarrow> a <: b ## M = (a <: M)" 
 Count_Remove_eq1   :   "a <: M -- a = (a <: M) - 1"
 Count_Remove_eq2   :   "a \<noteq> b \<Longrightarrow> a <: M -- b = (a <: M)" 
 Ext                :   "(\<And> a. (a <: M) = (a <: T)) \<Longrightarrow> M = T"


subsection {* Fold for bags *}


consts
 foldBag :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> 'a bag \<Rightarrow> 'b"


constdefs
 leftCommuting :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) set"
"leftCommuting \<equiv> {f. (\<forall> a b c. f a (f b c) = f b (f a c)) }" 

axioms
 foldBag1 : "foldBag f e {#} = e"
 foldBag2 : "f \<in> leftCommuting \<Longrightarrow> foldBag f e (x ## M) = f x (foldBag f e M)"





section {* Consequences *}

lemma Remove_Empty :
"{#} -- a = {#}"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(subst Count_Remove_eq1)
apply(subst Count_Empty)+
apply simp
apply(subst Count_Remove_eq2, assumption)
by (rule refl)

lemma Remove_idem :
"a <: M = 0 \<Longrightarrow> M -- a = M"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(subst Count_Remove_eq1)
apply clarsimp
apply(subst Count_Remove_eq2, assumption)
by (rule refl)


lemma Insert_Remove_eq1 :
"(a ## M) -- a = M"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(subst Count_Remove_eq1)
apply(subst Count_Insert_eq1)
apply simp
apply(subst Count_Remove_eq2, assumption)
apply(subst Count_Insert_eq2, assumption)
by(rule refl)


lemma Insert_Remove_eq2 :
"a \<noteq> b \<Longrightarrow> (a ## M) -- b = a ## M -- b"
apply(rule Ext)
apply(case_tac "aa = b", clarify)
apply(subst Count_Remove_eq1)
apply(subst Count_Insert_eq2, simp)+
apply(subst Count_Remove_eq1)
apply(rule refl)
apply(subst Count_Remove_eq2, assumption)
apply(case_tac "aa = a", clarify)
apply(subst Count_Insert_eq1)+
apply(subst Count_Remove_eq2, assumption)
apply(rule refl)
apply(subst Count_Insert_eq2, assumption)+
apply(subst Count_Remove_eq2, assumption)
by(rule refl)


lemma Insert_Remove_eq3 :
"0 < (a <: M) \<Longrightarrow> a ## M -- a = M"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(subst Count_Insert_eq1)
apply(subst Count_Remove_eq1)
apply simp
apply(subst Count_Insert_eq2, assumption)
apply(subst Count_Remove_eq2, assumption)
by(rule refl)


corollary Insert_Remove_eq4 :
"a <: M = 0 \<Longrightarrow> a ## M -- a = a ## M"
apply(subst Remove_idem, assumption)
by(rule refl)


lemma Insert_commute :
"a ## (b ## M) = b ## (a ## M)"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(subst Count_Insert_eq1)
apply(case_tac "a = b", clarify)
apply(subst Count_Insert_eq1)+
apply(rule refl)
apply(subst Count_Insert_eq2, assumption)+
apply(subst Count_Insert_eq1)
apply(rule refl)
apply(subst Count_Insert_eq2, assumption)
apply(case_tac "aa = b", clarify)
apply(subst Count_Insert_eq1)+
apply(subst Count_Insert_eq2, assumption)
apply(rule refl)
apply(subst Count_Insert_eq2, assumption)+
by(rule refl)


lemma Remove_commute :
"(M -- a) -- b = (M -- b) -- a"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(subst Count_Remove_eq1)
apply(case_tac "a = b", clarify)
apply(subst Count_Remove_eq1)+
apply(rule refl)
apply(subst Count_Remove_eq2, assumption)+
apply(subst Count_Remove_eq1)
apply(rule refl)
apply(subst Count_Remove_eq2, assumption)
apply(case_tac "aa = b", clarify)
apply(subst Count_Remove_eq1)+
apply(subst Count_Remove_eq2, assumption)
apply(rule refl)
apply(subst Count_Remove_eq2, assumption)+
by(rule refl)




section {* Cardinality *} 

constdefs
 cardinality :: "'a bag \<Rightarrow> nat"
"cardinality \<equiv> foldBag (\<lambda>a b. 1 + b) 0"

lemma cardinality0 : "cardinality {#} = 0"
apply(simp add: cardinality_def)
apply(subst foldBag1)
by(rule refl)

lemma cardinality1 : "cardinality (a ## M) = cardinality M + 1"
apply(simp add: cardinality_def)
apply(subst foldBag2)
apply(simp add: leftCommuting_def)
by(rule refl)


subsection {* Induction *}

lemma not_Empty_Ex : 
"S \<noteq> {#} \<Longrightarrow> \<exists>a. 0 < (a <: S)"
apply(rule ccontr, erule notE)
apply(rule Ext)
apply(subst Count_Empty)
by simp


lemma Remove_induct[rule_format] :
"P {#} \<Longrightarrow> \<forall>a M. 0 < (a <: M) \<longrightarrow> P (M -- a) \<longrightarrow> P M \<Longrightarrow> P X"
apply(rule_tac f=cardinality in measure_induct)
apply(case_tac "x = {#}", clarify)
apply(drule not_Empty_Ex)
apply clarify
apply(drule spec, drule spec, drule mp, assumption)
apply(erule mp)
apply(drule spec, erule mp)
apply(drule  Insert_Remove_eq3)
apply(erule_tac P="\<lambda>u. ?a < cardinality u" in subst)
apply(subst cardinality1)
apply simp
done



text {* FoldBag and removing *}

lemma foldBag3 :
"0 < (a <: S) \<Longrightarrow> f \<in> leftCommuting \<Longrightarrow> 
foldBag f e S = f a (foldBag f e (S -- a))"
apply(subst Insert_Remove_eq3[THEN sym], assumption)
apply(subst foldBag2, assumption)
apply(rule refl)
done



section {* Union *}


constdefs
 MUnion :: "'a bag \<Rightarrow> 'a bag \<Rightarrow> 'a bag"   ("_ \<squnion> _" [50, 51] 51)
"S \<squnion> T \<equiv> foldBag (\<lambda>x Y. x ## Y) T S" 



lemma MUnion_eq :
"a <: (S \<squnion> T) = (a <: S) + (a <: T)"
apply(induct_tac S rule: Remove_induct)
apply(subst MUnion_def)
apply(subst foldBag1)
apply(subst Count_Empty)
apply simp
apply(subst MUnion_def)
apply(subst foldBag3, assumption)
apply(simp add: leftCommuting_def)
apply(subst Insert_commute)
apply(simp)
apply(fold MUnion_def)
apply(case_tac "a = aa", clarify)
apply(subst Count_Insert_eq1)
apply simp
apply(subst Count_Remove_eq1)
apply simp
apply(subst Count_Insert_eq2, assumption)
apply simp
apply(subst Count_Remove_eq2, assumption)
by(rule refl)




section {* Map *}


constdefs
 Map :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a bag \<Rightarrow> 'b bag" 
"Map f \<equiv> foldBag (\<lambda>x Y. (f x) ## Y) {#}" 



lemma Map_le :
"(a <: M) \<le> (f a <: Map f M)"
apply(induct_tac M rule: Remove_induct)
apply(subst Map_def)
apply(subst foldBag1)
apply(subst Count_Empty)
apply simp
apply(subst Map_def)
apply(subst foldBag3, assumption)
apply(simp add: leftCommuting_def)
apply(subst Insert_commute)
apply(simp)
apply(fold Map_def)
apply(subst Insert_Remove_eq3[THEN sym], assumption)
apply(case_tac "a = aa", clarify)
apply(subst Count_Insert_eq1)+
apply simp
apply(subst Count_Insert_eq2, assumption)+
apply(erule le_trans)
apply(case_tac "f a = f aa", simp)
apply(subst Count_Insert_eq1)
apply simp
apply(subst Count_Insert_eq2, assumption)
by(rule le_refl)



lemma Map_inj :
"inj f \<Longrightarrow> (f a <: Map f M) \<le> (a <: M)"
apply(induct_tac M rule: Remove_induct)
apply(subst Map_def)
apply(subst foldBag1)
apply(subst Count_Empty)
apply simp
apply(subst Map_def)
apply(subst foldBag3, assumption)
apply(simp add: leftCommuting_def)
apply(subst Insert_commute)
apply(simp)
apply(fold Map_def)
apply(case_tac "f a = f aa", simp)
apply(subgoal_tac "a = aa", clarify)
apply(subst Count_Insert_eq1)
apply(rule le_trans)
apply(erule add_mono)
apply(rule le_refl)
apply(subst Count_Remove_eq1)
apply simp
apply(drule injD, assumption, clarify)
apply(subgoal_tac "a \<noteq> aa")
apply(subst Count_Insert_eq2, assumption)
apply(erule le_trans)
apply(subst Count_Remove_eq2, assumption)
apply(rule le_refl)
by force




end
