(*
   Title  : DList.thy
   Author : M. Bortin
*)

header {* The type of distinct lists as a quotient over lists *}


theory DList
imports "../../Extensions/AWE_HOL" 
begin

consts
 Norm :: "'a list \<Rightarrow> 'a list"
primrec
"Norm [] = []"
"Norm (x#xs) = x#(removeAll x (Norm xs))"

lemma removeAll_commute :
"removeAll a (removeAll b xs) = removeAll b (removeAll a xs)"
apply(induct_tac xs, simp_all)
apply force
done

lemma removeAll_distinct[rule_format, simp] :
"distinct xs \<longrightarrow> distinct (removeAll x xs)"
by(induct_tac xs, simp_all)

lemma Norm_distinct : 
"distinct (Norm xs)"
by(induct_tac xs, simp_all)

lemma removeAll_not_mem :
"\<not> a mem removeAll a xs" 
apply(induct_tac xs, simp_all)
apply clarify
done

lemma removeAll_mem :
"a \<noteq> b \<Longrightarrow> (a mem xs) = (a mem removeAll b xs)" 
apply(induct_tac xs, simp_all)
apply force
done

lemma Norm_mem :
"(x mem xs) = (x mem (Norm xs))"
apply(induct_tac xs, simp_all)
apply(case_tac "x=a", simp)
apply simp
apply(erule removeAll_mem)
done



lemma Norm_Nil[simp] :
"(Norm xs = []) = (xs = [])"
by(induct_tac xs, simp_all)

lemma Norm_hd[rule_format] :
" \<forall>ys. Norm xs = Norm ys \<longrightarrow> xs \<noteq> [] \<longrightarrow> hd xs = hd ys"
apply(induct_tac xs, simp)
apply clarsimp
apply(case_tac ys, simp)
by clarsimp



lemma Norm_remove_commute[simp] :
"Norm (removeAll x xs) = removeAll x (Norm xs)"
apply(induct_tac xs, simp_all)
apply(rule conjI, clarsimp)
apply(rule impI)
apply(rule removeAll_commute)
done


lemma removeAll_fix[rule_format] :
"\<not> x mem xs \<longrightarrow> removeAll x xs = xs" 
by(induct_tac xs, simp_all)

lemma removeAll_idem[simp] :
"removeAll x (removeAll x xs) = removeAll x xs"
apply(subst removeAll_fix)
apply(rule removeAll_not_mem)
by(rule refl)


lemma Norm_idem[simp] :
"Norm(Norm xs) = Norm xs"
by(induct_tac xs, simp_all)


lemma removeAll_length :
"length (removeAll x xs) \<le> length xs"
apply(induct_tac xs, simp_all)
by force


section {* Building quotient by the kernel of @{term "Norm"} *}

constdefs
 NormR :: "'a list \<times> 'a list \<Rightarrow> bool" 
"NormR \<equiv> {(xs, ys). Norm xs = Norm ys}"

syntax
 NormR_ :: "'a list \<Rightarrow> 'a list \<Rightarrow> bool"   ("_/ ~ _"  [50, 51] 50)
translations
"a ~ b" == "(a, b) \<in> NormR"


thymorph dlist_param : QuotientType_Parameters \<longrightarrow> DList
type_map: ["'a T" \<mapsto> "'a list"]
op_map: ["QuotientType_Parameters.eqR" \<mapsto> "DList.NormR"]


instantiate_theory QuotientType by_thymorph dlist_param
renames: [("QuotientType.TQ" \<mapsto> DList),
          ("QuotientType.qtype_ext" \<mapsto> extD mixfix: ("_~" [99] 1000)),
          ("QuotientType.repr_of" \<mapsto> repr_ofD mixfix: ("^^_" [99] 1000)),
          ("QuotientType.class_of" \<mapsto> class_ofD mixfix: ("\<guillemotleft>_\<guillemotright>" [99] 1000))
         ]
without_syntax

rename_thymorph dlist_param' to dlist
translate_thm "QuotientType.ext1" along dlist
translate_thm "QuotientType.ext2" along dlist
translate_thm "QuotientType.class_of_eq" along dlist
translate_thm "QuotientType.class_of_repr" along dlist
translate_thm "QuotientType.repr_of_class" along dlist
translate_thm "QuotientType.quotient_induct" along dlist



subsection{* Membership *}

lemma mem_congr : 
"congruence (op mem x)"
apply(unfold congruence_def)
apply clarify
apply(simp add: NormR_def)
apply(subst Norm_mem)
apply(erule ssubst)
apply(subst Norm_mem[THEN sym])
by(rule refl)

constdefs
 is_in :: "'a \<Rightarrow> 'a DList \<Rightarrow> bool"
"is_in x \<equiv> extD (op mem x)" 

lemma is_in_eq : 
"is_in x \<guillemotleft>xs\<guillemotright> = x mem xs"
apply(unfold is_in_def)
apply(subst ext1, rule mem_congr)
by(rule refl)


subsection {* Put *}

lemma Cons_congr :
"congruence (class_ofD \<circ> (Cons x))"
apply(unfold congruence_def)
apply clarsimp
apply(rule class_of_eq[THEN iffD1])
apply(simp add: NormR_def)
done


constdefs
 put :: "'a \<Rightarrow> 'a DList \<Rightarrow> 'a DList"
"put x \<equiv> extD (class_ofD \<circ> (Cons x))"

lemma put_eq : 
"put x \<guillemotleft>xs\<guillemotright> = \<guillemotleft>(x # xs)\<guillemotright>"
apply(unfold put_def)
apply(subst ext2, rule Cons_congr)
by(rule refl)


lemma is_in_put_eq :
"is_in x (put y ds) = (x = y \<or> is_in x ds)"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst put_eq)
apply(subst is_in_eq)
apply simp
apply(subst is_in_eq[THEN sym])
apply(subst class_of_repr)
by(rule refl)



lemma put_idem :
"put x (put x ds) = put x ds"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst put_eq)+
apply(rule trans)
prefer 2
apply(subst class_of_repr[THEN sym, of ds])
apply(rule refl)
apply(subst put_eq)
apply(rule class_of_eq[THEN iffD1])
apply(simp add: NormR_def)
done



subsection {* Empty *}

constdefs
 Empty :: "'a DList"
"Empty \<equiv> class_ofD []"

lemma put_not_Empty :
"put x ds \<noteq> Empty"
apply(rule notI)
apply(subst (asm) class_of_repr[THEN sym, of ds])
apply(subst (asm) put_eq)
apply(unfold Empty_def)
apply(drule class_of_eq[THEN iffD2])
apply(simp add: NormR_def)
done


lemma not_in_Empty :
"\<not> is_in x Empty"
apply(rule notI)
apply(unfold Empty_def)
apply(subst (asm) is_in_eq)
by simp


subsection {* head *}

lemma hd_congr :
"xs ~ ys \<Longrightarrow> \<not> (xs ~ []) \<Longrightarrow> hd xs = hd ys"
apply(simp add: NormR_def)
apply(rule Norm_hd, assumption)
apply(drule sym)
apply(rule notI, simp)
done


constdefs
 head :: "'a DList \<Rightarrow> 'a"
"head \<equiv> extD hd"


lemma head_eq :
"xs \<noteq> [] \<Longrightarrow> head \<guillemotleft>xs\<guillemotright> = hd xs"
apply(unfold head_def)
apply(unfold qtype_ext_def)
apply simp
apply(rule trans)
apply(rule hd_congr)
apply(rule repr_of_class)
apply(rule notI)
apply(subst (asm) class_of_eq)
apply(subst (asm) class_of_repr)
apply(subst (asm) class_of_eq[THEN sym])
apply(simp add: NormR_def)
by(rule refl)

lemma head_put :
"head (put x ds) = x"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst put_eq)
apply(subst head_eq)
by simp_all

lemma head_decomp1 :
"ds \<noteq> Empty \<Longrightarrow> ds = \<guillemotleft>(head ds # tl (^^ds))\<guillemotright>"
apply(subst (asm) class_of_repr[THEN sym, of ds])
apply(unfold Empty_def)
apply(subgoal_tac "^^ds \<noteq> []")
apply(drule hd_Cons_tl)
apply(subst class_of_repr[THEN sym, of ds], subst head_eq)
apply(rule notI, simp)
apply(erule ssubst)
apply(subst class_of_repr)
apply(rule refl)
apply clarsimp
done


lemma head_is_in :
"ds \<noteq> Empty \<Longrightarrow> is_in (head ds) ds"
apply(drule head_decomp1) 
apply(erule ssubst)
apply(subst is_in_eq)
apply(subst head_eq)
apply simp
by simp



subsection {* get *}

lemma rem_congr :
"congruence (class_ofD \<circ> removeAll x)"
apply(unfold congruence_def)
apply clarsimp
apply(rule class_of_eq[THEN iffD1])
apply(simp add: NormR_def)
done


constdefs
 get :: "'a \<Rightarrow> 'a DList \<Rightarrow> 'a DList"
"get x \<equiv> extD (class_ofD \<circ> removeAll x)"

lemma get_eq :
"get x \<guillemotleft>xs\<guillemotright> = \<guillemotleft>removeAll x xs\<guillemotright>"
apply(unfold get_def)
apply(subst ext2, rule rem_congr)
by(rule refl)


lemma not_is_in_get :
"\<not> is_in x (get x ds)"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst get_eq)
apply(subst is_in_eq)
by (rule removeAll_not_mem)

lemma is_in_get :
"a \<noteq> b \<Longrightarrow> is_in a (get b ds) = is_in a ds"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst get_eq)
apply(subst is_in_eq)
apply(subst removeAll_mem[THEN sym], assumption)
apply(subst is_in_eq[THEN sym])
apply(subst class_of_repr)
by(rule refl)


lemma head_decomp2 :
"ds \<noteq> Empty \<Longrightarrow> ds = put (head ds) (get (head ds) ds)"
apply(drule head_decomp1)
apply(erule ssubst)
apply(subst head_eq, simp)+
apply simp
apply(subst get_eq)
apply simp
apply(subst put_eq)
apply(subst class_of_eq[THEN sym])
apply(simp add: NormR_def)
done

lemma get_fix :
"\<not> is_in x ds \<Longrightarrow> get x ds = ds"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst get_eq)
apply(subst (asm) class_of_repr[THEN sym, of ds])
apply(subst (asm) is_in_eq)
apply(subst removeAll_fix, assumption)
apply(subst class_of_repr)
by(rule refl)

lemma get_put_eq :
"\<not> is_in x ds \<Longrightarrow> get x (put x ds) = ds"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst put_eq)
apply(subst get_eq)
apply simp
apply(subst get_eq[THEN sym])
apply(subst class_of_repr)
by(erule get_fix)

lemma get_Empty :
"get x Empty = Empty"
apply(unfold Empty_def)
apply(subst get_eq)
by simp


lemma put_head :
"ds \<noteq> Empty \<Longrightarrow> put (head ds) ds = ds"
apply(drule head_decomp2)
apply(rule_tac P="\<lambda>x. put ?a x = ?b" in ssubst)
apply assumption
apply(subst put_idem)
by simp


lemma get_put_commute :
"a \<noteq> b \<Longrightarrow> get a (put b ds) = put b (get a ds)"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst put_eq)
apply(subst get_eq)
apply simp
apply(subst put_eq[THEN sym])
apply(subst get_eq[THEN sym])
apply(subst class_of_repr)
by(rule refl)


corollary get_put_commute_head :
"ds \<noteq> Empty \<Longrightarrow> a \<noteq> head ds \<Longrightarrow> get a ds = put (head ds) (get a (get (head ds) ds))"
apply(drule head_decomp2)
apply(erule_tac P="\<lambda>x. get a x = ?u" in ssubst)
apply(subst get_put_commute, assumption)
by(rule refl)

lemma get_commute :
"get a (get b ds) = get b (get a ds)"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst get_eq)+
apply(subst removeAll_commute)
apply(subst get_eq[THEN sym])+
apply(subst class_of_repr)
by(rule refl)



subsection {* Fold *}


constdefs
 FoldD :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> 'a DList \<Rightarrow> 'b"
"FoldD f e \<equiv> extD (\<lambda>xs. foldr f (Norm xs) e)" 


lemma FoldD_eq :
"FoldD f e \<guillemotleft>xs\<guillemotright> = foldr f (Norm xs) e"
apply(unfold FoldD_def)
apply(subst ext1)
apply(simp add: congruence_def NormR_def)
by(rule refl)

lemma FoldD_eq1 :
"FoldD f e Empty = e"
apply(unfold Empty_def)
apply(subst FoldD_eq)
by simp

lemma FoldD_eq2 :
"\<not> is_in x ds \<Longrightarrow> FoldD f e (put x ds) = f x (FoldD f e ds)"
apply(subst class_of_repr[THEN sym, of ds])
apply(subst put_eq)
apply(subst FoldD_eq)
apply simp
apply(subst (asm) class_of_repr[THEN sym, of ds])
apply(subst (asm) is_in_eq)
apply(subst removeAll_fix)
apply(subst Norm_mem[THEN sym])
apply(assumption)
apply(simp add: FoldD_def qtype_ext_def)
done




corollary FoldD_put_head :
"FoldD f e (put (head ds) (get (head ds) ds)) = f (head ds) (FoldD f e (get (head ds) ds))"
apply(rule FoldD_eq2)
apply(rule not_is_in_get)
done

corollary FoldD_get : 
"ds \<noteq> Empty \<Longrightarrow> FoldD f e ds = f (head ds) (FoldD f e (get (head ds) ds))"
apply(subst head_decomp2[of ds], assumption)
apply(subst FoldD_put_head)
by(rule refl)



section {* Induction *}

lemma DList_induct1[rule_format]: 
"P Empty \<Longrightarrow> \<forall>x ds. P ds \<longrightarrow> P (put x ds) \<Longrightarrow> P ds"
apply(rule quotient_induct)
apply(induct_tac a)
apply(fold Empty_def, assumption)
apply(subst put_eq[THEN sym])
by force

lemma DList_induct2[rule_format]: 
"P Empty \<Longrightarrow> \<forall>ds. ds \<noteq> Empty \<longrightarrow> P (get (head ds) ds) \<longrightarrow> P ds \<Longrightarrow> P ds"
apply(rule quotient_induct)
apply(rule_tac f="length" in measure_induct)
apply(case_tac x, clarify)
apply(fold Empty_def, assumption)
apply clarify
apply(drule_tac x="\<guillemotleft>(aa#list)\<guillemotright>" in spec)
apply(drule mp)
apply(subst put_eq[THEN sym])
apply(rule put_not_Empty)
apply(erule mp)
apply(subst head_eq, simp)
apply simp
apply(subst get_eq)
apply(drule spec, erule mp)
apply simp
apply(rule le_less_trans)
apply(rule removeAll_length)
by simp



theorem FoldD_left_commute[rule_format] :
"(\<forall> a b c. f a (f b c) = f b (f a c)) \<Longrightarrow> 
 is_in x ds \<longrightarrow> FoldD f e (put x (get x ds)) = FoldD f e ds"
apply(rule DList_induct2)
apply(simp add: not_in_Empty)
apply clarify
apply(case_tac "x = head ds", clarify)
apply(subst head_decomp2[THEN sym], assumption)
apply(rule refl)
apply(frule head_decomp2)
apply(erule_tac ssubst)
apply(subst get_put_commute, assumption)
apply(subst FoldD_eq2)
apply(subst is_in_put_eq)
apply simp
apply(rule not_is_in_get)
apply(subst FoldD_eq2)
apply(subst is_in_get, simp)
apply(rule not_is_in_get)
apply(drule spec)+
apply(erule subst)
apply(subst FoldD_eq2[where f=f, THEN sym])
apply(rule not_is_in_get)
apply(drule mp)
apply(subst is_in_get, simp)
apply assumption
apply(erule ssubst)
apply(subst FoldD_put_head)
by(rule refl)



end
