(*
   Title  : FiniteSet.thy
   Author : M. Bortin
*)

header {* Implementation of finite sets w.r.t. FiniteSet-Spec *}


theory FiniteSet
imports DList
begin

constdefs
 eqR :: "'a DList \<times> 'a DList \<Rightarrow> bool"
"eqR \<equiv> (\<lambda>(xs, ys). (\<forall>x. is_in x xs = is_in x ys))"


sigmorph fset : QuotientType_Parameters \<longrightarrow> FiniteSet
type_map: ["'a T" \<mapsto> "'a DList"]

thymorph_goals fset


lemma eqR_sym:
"a \<simeq> b \<Longrightarrow> b \<simeq> a"
by(simp add: eqR_def mem_def)

lemma eqR_refl:
"a \<simeq> a"
by(simp add: eqR_def mem_def)

lemma eqR_trans:
"\<lbrakk>a \<simeq> b; b \<simeq> c\<rbrakk> \<Longrightarrow> a \<simeq> c"
by(simp add: eqR_def mem_def)

thymorph fset by_sigmorph fset

instantiate_theory QuotientType by_thymorph fset
renames: [("QuotientType.TQ" \<mapsto> FSet)]


translate_thm "QuotientType.ext1" along fset'
translate_thm "QuotientType.ext2" along fset'
translate_thm "QuotientType.class_of_eq" along fset'
translate_thm "QuotientType.class_of_repr" along fset'


subsection {* Some basic properties of the equivalence relation on distinct lists *}

lemma eqR_eq : "(s \<simeq> t) = (\<forall>x. (is_in x s) = (is_in x t))"
by(simp add: eqR_def mem_def)

lemma eqR_Empty : "s \<simeq> Empty \<Longrightarrow> s = Empty"
apply(subst (asm) eqR_eq)
apply(rule ccontr)
apply(drule head_is_in)
apply(drule_tac x="head s" in spec)
apply(simp add: not_in_Empty)
done



lemma eqR_is_in_head :
"\<lbrakk> s \<simeq> t; s \<noteq> Empty \<rbrakk> \<Longrightarrow> is_in (head s) t"
apply(subst (asm) eqR_eq)
apply(drule head_is_in)
apply(simp)
done



section {* Deriving fold for finite sets *}


instantiate_theory Fold_Parameters by_thymorph fset'
type_map: ["('u, 'a) \<Sigma>" \<mapsto> "unit + 'u \<times> 'a"]



defs
 TC_def : "TC ds \<equiv> (if ds = Empty then Inl () 
                    else Inr (head ds, get (head ds) ds))"
 

 Sigma_def : "\<Sigma> f q \<equiv> (case q of 
                        Inl u \<Rightarrow> Inl u
                      | Inr (x, xs) \<Rightarrow> Inr (x, f xs))"

 Sigma'_def : "\<Sigma>' X \<equiv> {(Inl (), Inl ())} \<union> {(Inr (a, xs), Inr (a, ys)) | a xs ys. (xs, ys) \<in> X}"


 Transform_def : "Transform s t \<equiv> (case t of 
                                    Inl u \<Rightarrow> Inl u
                                  | Inr (x, xs) \<Rightarrow> (case s of
                                                     Inl u \<Rightarrow> Inr (x, xs)
                                                   | Inr (y, ys) \<Rightarrow> Inr (y, get y (put x xs))))"
 FoldCond_def : "FoldCond \<equiv> UNIV" 

 TransformCond_def : "TransformCond \<equiv> {A. \<forall>a b c. A (Inr (a, A (Inr (b, c)))) = A (Inr (b, A (Inr (a, c))))}"

 Fold_def : "Fold A \<equiv> FoldD (\<lambda>a b. A (Inr (a, b))) (A (Inl ()))"  



subsection {* Proving the theory morphism goals for the signature morphism fset'' : Fold1 --> FiniteSet *}

lemma TransformCond:
"TransformCond \<subseteq> FoldCond"
by(simp add: FoldCond_def)


lemma Fold_eq:
"A \<in> FoldCond \<Longrightarrow> Fold A = A \<circ> \<Sigma> (Fold A) \<circ> TC"
apply(rule ext, simp)
apply(case_tac "x = Empty", clarify)
apply(simp add: TC_def)
apply(simp add: Sigma_def)
apply(simp add: Fold_def)
apply(subst FoldD_eq1)
apply(rule refl)
apply(simp add: TC_def)
apply(simp add: Sigma_def)
apply(subst Fold_def)
apply(subst FoldD_get)
apply(assumption)
apply(fold Fold_def)
by(rule refl)

lemma TC_reduct:
"TC -` \<Sigma>'' P \<subseteq> P \<Longrightarrow> UNIV \<subseteq> P"
apply(subst (asm) subset_iff)
apply clarify
apply(induct_tac x rule: DList_induct2)
apply(drule spec, erule mp)
apply(simp add: TC_def)
apply(simp add: Sigma''_def monoP_def)
apply(simp add: Sigma'_def)
apply(drule spec, erule mp)
apply(simp add: TC_def)
apply(simp add: Sigma''_def monoP_def)
apply(simp add: Sigma'_def)
apply(simp add: monoE_def)
done



lemma Transform_eqR :
"s \<simeq> t \<Longrightarrow> TC s \<simeq>\<^sub>\<Sigma> Transform (TC s) (TC t)"
apply(case_tac "s = Empty", clarify)
apply(drule eqR_sym)
apply(drule eqR_Empty)
apply clarify
apply(simp add: TC_def Transform_def eqR_ext_def)
apply(simp add: Sigma'_def)
apply(subgoal_tac "t \<noteq> Empty")
apply (simp add: TC_def Transform_def)
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply(subst eqR_eq)
apply(rule allI)
apply(case_tac "x= head s", clarify)
apply(simp add: not_is_in_get)
apply(simp add: is_in_get)
apply(case_tac "x = head t", clarify)
apply(subst is_in_put_eq, simp)
apply(rule eqR_is_in_head)
apply(erule eqR_sym, assumption)
apply(simp add: is_in_put_eq)
apply(subst (asm) eqR_eq)
apply(drule_tac x=x in spec)
apply(subst is_in_get)
apply(assumption+)
apply(clarify)
apply(drule eqR_Empty)
by clarify


lemma Sigma_division:
"\<Sigma>'' (eqR \\ ker f) \<subseteq> eqR_ext \\ ker (\<Sigma> f)"
apply clarify
apply(case_tac x, clarify)
apply(simp add: division_def)
apply clarify
apply(case_tac b, clarify)
apply(simp add: ker_def)
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply clarify
apply(simp add: division_def)
apply clarify
apply(case_tac ba, clarify)
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply clarify
apply(simp add: ker_def)
apply(simp add: eqR_ext_def)
apply(simp add: Sigma'_def)
apply clarify
apply(simp add: Sigma''_def monoP_def)
apply(simp add: Sigma'_def monoE_def)
apply(simp add: Sigma_def)
done




lemma TransformCond_subst :
"A \<in> TransformCond \<Longrightarrow> A (Inr (a, A (Inr (b, c)))) = A (Inr (b, A (Inr (a, c))))"
by(simp add: TransformCond_def)


lemma Fold_Transform:
"\<lbrakk>A \<in> TransformCond; s \<simeq> t\<rbrakk>
\<Longrightarrow> A (\<Sigma> (Fold A) (Transform (TC s) (TC t))) = Fold A t"
apply(rule sym)
apply(subst Fold_eq, rule subsetD, rule TransformCond, assumption)
apply(rule sym, simp)
apply(case_tac "s = Empty", clarify)
apply(drule eqR_sym)
apply(drule eqR_Empty)
apply clarify
apply(simp add: TC_def Transform_def eqR_ext_def)
apply(subgoal_tac "t \<noteq> Empty")
apply(simp add: TC_def)
apply(simp add: Transform_def)
apply(simp add: Sigma_def)
apply(subst head_decomp2[THEN sym], assumption)
apply(subst Fold_def)
apply(subst FoldD_eq2[where f="\<lambda>a b. A (Inr (a, b))", THEN sym])
apply(rule not_is_in_get)
apply(subst FoldD_left_commute)
apply(subst TransformCond_subst[of A], assumption)
apply(rule refl)
apply(rule eqR_is_in_head, assumption+)
apply(fold Fold_def)
apply(subst Fold_def, subst FoldD_get[where f="\<lambda>a b. A (Inr (a, b))", THEN sym], assumption)
apply(fold Fold_def)
apply(rule refl)
apply clarify
apply(drule eqR_Empty)
by clarify


thymorph fold_fset by_sigmorph fset''

instantiate_theory Fold by_thymorph fold_fset
renames: [("Fold.FoldQ" \<mapsto> FoldSet),
          ("Fold.FoldQ_def" \<mapsto> FoldSet_def)]

translate_thm "Fold.FoldQ_eq" as FoldSet_eq along fold_fset'




section {* Proving an instance of FiniteSet_Spec *}

thymorph i : Nat \<longrightarrow> FiniteSet (* an inclusion *)

instantiate_theory FiniteSet_Spec by_thymorph i
type_map: ["'a finite_set" \<mapsto> "'a FSet"] 
renames: [("FiniteSet_Spec.Empty" \<mapsto> Empty_FSet mixfix: ("{#}"))] 


defs
 Empty_FSet_def   : "{#} \<equiv> <Empty>"

 Insert_def       : "Insert a \<equiv> (class_of \<circ> put a)\<^sup>#"

 Mem_def          : "Mem a \<equiv> (is_in a)\<^sup>#"

 Remove_def       : "(S -- a) \<equiv>  (class_of \<circ> get a)\<^sup># S"

 foldSet_def      : "foldSet f e \<equiv> FoldSet (\<lambda>x. case x of Inl u \<Rightarrow> e | Inr (a, b) \<Rightarrow> f a b) "



subsection {* Proving congruences for the basic operations *}

lemma is_in_congr :
"congruence (is_in x)"
apply(simp add: congruence_def)
apply(clarify)
apply(subst (asm) eqR_eq)
by simp

lemma Mem_eq :
"a <: <ds> = is_in a ds"
apply(unfold Mem_def)
apply(subst ext1, rule is_in_congr)
by(rule refl)


lemma put_congr : "congruence (class_of \<circ> put x)"
apply(simp add: congruence_def)
apply(clarify)
apply(subst class_of_eq[THEN sym])
apply(subst eqR_eq)
apply clarify
apply(subst is_in_put_eq)+
apply(case_tac "xa = x", simp)
apply simp
apply(subst (asm) eqR_eq)
apply force
done

lemma Insert_eq :
"a ## <ds> = <put a ds>"
apply(unfold Insert_def)
apply(subst ext2, rule put_congr)
by(rule refl)


lemma get_congr :
"congruence (class_of \<circ> get x)"
apply(simp add: congruence_def)
apply clarify
apply(subst class_of_eq[THEN sym])
apply(subst eqR_eq)
apply(rule allI)
apply(case_tac "xa = x", clarify)
apply(simp add: not_is_in_get)
apply(subst is_in_get, assumption)+
apply(subst (asm) eqR_eq)
by simp

lemma Remove_eq :
"(<ds> -- a) = <get a ds>" 
apply(simp add: Remove_def)
apply(subst ext2, rule get_congr)
by(rule refl)


subsection {* Proving the theory morphism goals *}


lemma Ext:
"(\<And>a. (a <: S) = (a <: T)) \<Longrightarrow> S = T"
apply(subst (asm) class_of_repr[THEN sym, of S])
apply(subst (asm) class_of_repr[THEN sym, of T])
apply(subst (asm) Mem_eq)+
apply(drule eqR_eq[THEN iffD2, rule_format])
apply(subst (asm) class_of_eq)
apply(subst (asm) class_of_repr)+
by assumption

lemma Mem_Remove_eq:
"(a <: S -- b) = (a \<noteq> b \<and> a <: S)"
apply(subst class_of_repr[THEN sym, of S])
apply(subst Remove_eq)
apply(subst Mem_eq)
apply(case_tac "a = b", simp add: not_is_in_get)
apply(subst is_in_get, assumption)
apply(subst Mem_eq[THEN sym])
apply(subst class_of_repr)
by simp


lemma Mem_Insert_eq:
"(a <: b ## S) = (a = b \<or> a <: S)"
apply(subst class_of_repr[THEN sym, of S])
apply(subst Insert_eq)
apply(subst Mem_eq)
apply(subst is_in_put_eq)
apply(subst Mem_eq[THEN sym])
apply(subst class_of_repr)
by(rule refl)


lemma not_Mem_Empty:
"\<not> a <: {#}"
apply(unfold Empty_FSet_def)
apply(subst Mem_eq)
by(rule not_in_Empty)



lemma foldSet1:
"foldSet f e {#} = e"
apply(unfold Empty_FSet_def foldSet_def)
apply(unfold FoldSet_def)
apply(simp add: qtype_ext_def)
apply(subst eqR_Empty[of "^<Empty>"])
apply(rule class_of_eq[THEN iffD2])
apply(subst class_of_repr)
apply(rule refl)
apply(subst Fold_def)
apply(subst FoldD_eq1)
by simp


lemma foldSet2:
"\<lbrakk>f \<in> leftCommuting; \<not> x <: S\<rbrakk> \<Longrightarrow> foldSet f e (x ## S) = f x (foldSet f e S)"
apply(subst class_of_repr[THEN sym, of S])
apply(subst Insert_eq)
apply(subst foldSet_def)
apply(subst FoldSet_eq)
apply(simp add: TransformCond_def leftCommuting_def)
apply(unfold Fold_def)
apply(subst FoldD_eq2)
apply(subst Mem_eq[THEN sym])
apply(subst class_of_repr)
apply assumption
apply(fold Fold_def)
apply(subst FoldSet_eq[THEN sym])
apply(simp add: TransformCond_def leftCommuting_def)
apply simp
apply(fold foldSet_def)
apply(subst class_of_repr)
by(rule refl)


thymorph finite_set_spec by_sigmorph i'

translate_thm "FiniteSet_Spec.FinitePred_induct" along finite_set_spec
translate_thm "FiniteSet_Spec.FinitePred_Inter" along finite_set_spec





end


