(*
 * Theory morphisms
 * $Id: thymorph.sml 1191 2010-01-20 11:57:49Z maxim $
 *)

signature THYMORPH =
sig
 type thymorph
 datatype ThyMorphData = ThyMorphData of 
                          { dom_thms : (Thm.thm * Term.term * Symtab.key list) Symtab.table,
                            cod_thms : (Thm.thm * Term.term) Symtab.table }

 val data : thymorph -> ThyMorphData
 val get_thymorph_data : Sigmorph.sigmorph -> ThyMorphData
 val sigmorph : thymorph -> Sigmorph.sigmorph
 val axiomMap : thymorph ->  Derive_Thm.AxmImg Symtab.table
 val theoremMap : thymorph -> Symtab.key Symtab.table
 val match_thms : thymorph -> thymorph
 val normalize : (string * mixfix option) Symtab.table -> thymorph -> thymorph
 val translate_syntax : (string * mixfix option) Symtab.table -> thymorph -> thymorph
 val translate_thm : thymorph -> (string * mixfix option) Symtab.table -> 
                     string -> (thymorph * ((string * Thm.thm) list))
 val compose : (string * mixfix option) Symtab.table -> thymorph -> thymorph -> thymorph * thymorph
 val cod_ext : thymorph -> Symtab.key Symtab.table -> theory -> thymorph

 val thymorph_cmd : bool -> theory -> theory -> 
            (string * string) list -> (string * string) list -> (string * string) list ->
            (string * string) list -> thymorph
 val thymorph_by_sigmorph_cmd : bool -> Sigmorph.sigmorph -> (string * string) list -> thymorph

 datatype Inst_Result = ThyM of thymorph 
                      | SigM of (Sigmorph.sigmorph * ((string * Thm.cterm) list))

 val instantiate : bool -> bool -> (string * string) list -> (string * string) list -> 
                  (string * string) list -> (string * string) list ->
                  (string * mixfix option) Symtab.table -> 
                   theory -> thymorph -> Inst_Result

 val thymorph_goals : bool -> (string * mixfix option) Symtab.table -> Sigmorph.sigmorph -> 
                      (Sigmorph.sigmorph * ((string * Thm.cterm) list)) 

 val add_thymorph : (string * thymorph) -> unit
 val get_thymorph : string -> thymorph option
 val print_thymorphs : unit -> string
 
 val thymorph_to_string : thymorph -> string

 exception THYMORPH of string

end


structure Thymorph : THYMORPH =
struct

 exception THYMORPH of string

 datatype ThyMorphData = ThyMorphData of 
                          { dom_thms : (Thm.thm * Term.term * Symtab.key list) Symtab.table,
                            cod_thms : (Thm.thm * Term.term) Symtab.table }




 datatype thymorph = ThyMorph of { sigma : Sigmorph.sigmorph,
                                   axiomMap : Derive_Thm.AxmImg Symtab.table,
                                   theoremMap : Symtab.key Symtab.table,
                                   data : ThyMorphData }



fun sigmorph (ThyMorph{ sigma, ... }) = sigma
fun axiomMap (ThyMorph{ axiomMap, ... }) = axiomMap
fun theoremMap (ThyMorph{ theoremMap, ...}) = theoremMap
fun data (ThyMorph{ data, ... }) = data


fun assoc_defs defnm_rel dep_ops = Aux.map_foldl (Symtab.lookup_list defnm_rel) Aux.union_str ([], dep_ops)


     fun ext_mapping mp [] = mp
       | ext_mapping mp ((x as (key, v))::xs) =
         let val mp' = Symtab.update_new x mp
                       handle DUP => (let val _ = writeln ("Updating " ^ key)
                                      in
                                       Symtab.update x mp
                                      end)
         in
          ext_mapping mp' xs
         end

   
     fun prepare_dom_thms dg t [] = t
       | prepare_dom_thms dg t ((nm, thm)::xs) =
         let val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
             val thm_dep = (case (Symtab.lookup dg nm) of
                             NONE => []
                           | SOME (_, deps) => deps)
         in
          prepare_dom_thms dg (Symtab.update (nm, (thm, norm thm, thm_dep)) t) xs
         end 
     

     fun prepare_cod_thms t [] = t
       | prepare_cod_thms t ((nm, thm)::xs) =
         let val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
         in
          prepare_cod_thms (Symtab.update (nm, (thm, norm thm)) t)xs
         end      




fun get_thymorph_data sigma =
   let val Sigmorph.SigMorphData d = Sigmorph.data sigma
       val src = Sigmorph.src sigma
       val trg = Sigmorph.trg sigma
       val src_id = Aux.name_of_thy src
       val trg_id = Aux.name_of_thy trg

       val dom = #dom d
       val cod = #cod d
       val global_thys = #global d
       val domain_thys = (if (dom = []) then [] else src::(map theory (remove1 (op =) src_id dom)))
       val codomain_thys = (if (cod = []) then [] else trg::(map theory (remove1 (op =) trg_id cod)))

       val dom_thms = Aux.map_foldl Aux.flat_thms_of op@ ([], domain_thys)
       val cod_thms = Aux.map_foldl Aux.flat_thms_of op@ ([], codomain_thys)
       val dom_thms_dep = Dependency.thms_dep_graph dom_thms
       val dom_thms' = prepare_dom_thms dom_thms_dep Symtab.empty dom_thms
       val cod_thms' = prepare_cod_thms Symtab.empty cod_thms

   in
    ThyMorphData { dom_thms = dom_thms', cod_thms = cod_thms' }
   end




(* **************************** *)
(* *** Theorem matching ******* *)
(* **************************** *)

fun match_thms_i sigma thmMap dom_thms cod_thms = 
  let val trg = Sigmorph.trg sigma
      val tsig = Sign.tsig_of trg

      fun match_thm tsig (nm, term) [] = NONE
        | match_thm tsig (nm, term) ((x as (nm', (thm', term')))::xs) =
            if (Alpha.term_aconv tsig (term, term')) 
            then (let val _ = Aux.writeln ("Found: (" ^ (quote nm) ^ " |-> " ^ (quote nm') ^ ")")
                  in
                   SOME (nm, nm')
                  end)
            else match_thm tsig (nm, term) xs
                  
      fun match_thms_ tsig ext cod_thms [] = ext
        | match_thms_ tsig ext cod_thms ((nm, (thm, term, _))::xs) =
          (case (Symtab.lookup ext nm) of
            NONE => (let val term' = (Thm.term_of o (Thm.cterm_of trg) o (Sigmorph.map_term sigma)) term
                     in
                       case (match_thm tsig (nm, term') cod_thms) of
                        NONE => match_thms_ tsig ext cod_thms xs
                      | SOME mp => (let val ext' = ext_mapping ext [mp]
                                    in
                                     match_thms_ tsig ext' cod_thms xs
                                    end)
                     end)
          | _ => match_thms_ tsig ext cod_thms xs)
  in
   match_thms_ tsig thmMap cod_thms dom_thms
  end   

fun match_thms (ThyMorph{sigma, axiomMap, theoremMap, data = ThyMorphData d}) =
   let val dom_thms = Symtab.dest (#dom_thms d)
       val cod_thms = Symtab.dest (#cod_thms d)      
       val theoremMap' = match_thms_i sigma theoremMap dom_thms cod_thms
   in
     ThyMorph{sigma = sigma, axiomMap = axiomMap, 
              theoremMap = theoremMap', data = ThyMorphData d}
   end


(* **************************** *)
(* *** Axiom matching ********* *)
(* **************************** *)



fun store_thm (nm, thm) thy =
 (case (try (PureThy.store_thm (Binding.name nm, thm)) thy) of
   SOME (_, thy') => (nm, thy')
 | NONE => store_thm (nm ^ "'", thm) thy)


fun match_axms all sigma axiomMap cod_thms =
 let val Sigmorph.SigMorphData d = Sigmorph.data sigma
     val dom_axms = Symtab.dest (#dom_axs_map d)
     val cod_axms = Symtab.dest (#cod_axs_map d)
     val trg = (Sigmorph.trg sigma)
     val trg_id = Aux.name_of_thy trg
     val tsig = Sign.tsig_of trg
     val cod_defs = #sorted_cod_defs d
     val opMap = Symtab.dest (Sigmorph.opMap sigma)
     val op_basis = map fst (#op_basis d)

     fun match_by_ax term [] = NONE
       | match_by_ax term ((nm, term')::xs) =
            (if (Alpha.term_aconv tsig (term, term')) then SOME nm 
             else match_by_ax term xs)
     fun match_by_thm term [] = NONE
       | match_by_thm term ((nm, (thm, term'))::xs) =
            (if (Alpha.term_aconv tsig (term, term')) then SOME nm 
             else match_by_thm term xs)
 

     fun match_ax thy axMap notmapped [] = (thy, axMap, notmapped)
       | match_ax thy axMap notmapped ((nm, term)::xs) =
       (case (Symtab.lookup axMap nm) of
         NONE => (let val term' = (Thm.term_of o (Thm.cterm_of thy) o (Sigmorph.map_term sigma)) term 
                  in
                   case (match_by_ax term' cod_axms) of
                    NONE => 
                     (case (match_by_thm term' cod_thms) of
                       NONE => (let val (q, bnm_) = (Aux.qname_split nm)
                                    val bnm = (q ^"_"^ bnm_)
                                in
                                 (case (Prove.proof_search thy op_basis opMap cod_defs bnm term') of
                                   NONE => match_ax thy axMap ((nm, term')::notmapped) xs
                                 | SOME thm_ => 
                                   (let val thm = (Drule.standard o Thm.varifyT o Aux.beta_eta_convert) thm_
                                        val (bnm', thy') = store_thm (bnm, thm) thy
                                        val _ = Aux.info ("Lemma " ^ bnm' ^" : "^
                                                         (Display.string_of_thm (ProofContext.init thy') thm) ^ 
                                                          " proven.")                                        
                                        val qnm = (trg_id ^ "." ^ bnm')
                                        val axMap' = Symtab.update_new (nm, Derive_Thm.Thm qnm) axMap
                                    in
                                       match_ax thy' axMap' notmapped xs
                                    end))
                                end)
                     | SOME i => (let val _ = Aux.info ("Axiom mapping found: (" ^ (quote nm) ^ 
                                                                " |-> " ^ (quote i) ^ ")")
                                  in
                                    match_ax thy (Symtab.update_new (nm, Derive_Thm.Thm i) axMap) notmapped xs
                                  end))
                  | SOME i => (let val _ = Aux.info ("Axiom mapping found: (" ^ (quote nm) ^ 
                                                      " |-> " ^ (quote i) ^ ")")
                               in
                                match_ax thy (Symtab.update_new (nm, Derive_Thm.Axm i) axMap) notmapped xs
                               end)
                  end)
      | SOME i => match_ax thy axMap notmapped xs)

     val (trg', axMap, notmapped) = match_ax trg axiomMap [] dom_axms 
     val sigma' = Sigmorph.cod_ext sigma trg'
 in
  if (not all) then (sigma', axMap)
  else 
   (case notmapped of
     [] => (sigma', axMap)
    | _ =>  (let fun show [] = ""
                   | show ((nm, term)::xs) = 
                     let val (q, b) = Aux.qname_split nm 
                     in
                       (q ^ "_" ^ b ^ " : " ^ quote(Syntax.string_of_term_global trg term) ^ "\n" ^ (show xs))
                     end  
                 val _ = Aux.error_msg ("Cannot find following proposition(s) in the codomain: \n" ^ 
                                        (show notmapped))
             in
              raise THYMORPH "Theory morphism construction failed."
             end))
 end 


(* **************************** *)
(* *** Construction     ******* *)
(* **************************** *)

fun transf_axMap [] = []
  | transf_axMap ((a, b)::xs) = (case b of
                                    Derive_Thm.Axm b' => (a, b')::(transf_axMap xs)
                                  | Derive_Thm.Thm b' => (a, b')::(transf_axMap xs))


fun thymorph_by_sigmorph_cmd norm sigma_ axMap_rep =
 let val sigma__ = Sigmorph.match_defs sigma_
     val sigma = (if norm then Sigmorph.normalize Symtab.empty sigma__ else sigma__)
     val Sigmorph.SigMorphData d = Sigmorph.data sigma
     val (td as (ThyMorphData { dom_thms, cod_thms })) = get_thymorph_data sigma

     val dom_axms = #dom_axs_map d
     val cod_axms = #cod_axs_map d
     val trg = Sigmorph.trg sigma
     val tsig = Sign.tsig_of trg

     fun mk_axiomMap err t [] = (err, t)
       | mk_axiomMap err t ((a, b)::xs) =
     (case (Symtab.lookup dom_axms a) of
      NONE => mk_axiomMap err t xs
    | SOME term => 
       (case (Symtab.lookup cod_axms b) of
         SOME term' => 
              (if (Alpha.term_aconv tsig (Sigmorph.map_term sigma term, term'))
               then mk_axiomMap err (Symtab.update (a, Derive_Thm.Axm b) t) xs
               else mk_axiomMap ((a, b)::err) t xs)
     | NONE => (case (Symtab.lookup cod_thms b) of
                 SOME (thm', term') => (if (Alpha.term_aconv tsig (Sigmorph.map_term sigma term, term'))
                                        then mk_axiomMap err (Symtab.update (a, Derive_Thm.Thm b) t) xs
                                        else mk_axiomMap ((a, b)::err) t xs) 
               | NONE => (let val pthm = (true, Aux.get_axiom trg b)
                                          handle e => (false, Aux.get_thm trg b)
                              val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
                          in
                           case pthm of
                            (true, thm') => (let val term' = norm thm'
                                             in
                                              (if (Alpha.term_aconv tsig (Sigmorph.map_term sigma term, term'))
                                               then mk_axiomMap err (Symtab.update (a, Derive_Thm.Axm b) t) xs
                                               else mk_axiomMap ((a, b)::err) t xs)
                                             end)
                         | (false, thm') => (let val term' = norm thm'
                                             in
                                              (if (Alpha.term_aconv tsig (Sigmorph.map_term sigma term, term'))
                                               then mk_axiomMap err (Symtab.update (a, Derive_Thm.Thm b) t) xs
                                               else mk_axiomMap ((a, b)::err) t xs)
                                             end)
                          end))))

     val (sigma', axiomMap) = 
                  (case (mk_axiomMap [] Symtab.empty axMap_rep) of
                      ([], t) => match_axms true sigma t (Symtab.dest cod_thms) 
                    | (err, t) => 
                    (let val _ = Aux.error_msg ("Following axiom mapping(s) are illegal:\n" ^ 
                                                 AWE.show_map_i err)
                     in
                       raise THYMORPH "Theory morphism construction failed."
                     end))  
     val tmorph =   ThyMorph { sigma = sigma', axiomMap = axiomMap, 
                               theoremMap = Symtab.empty, data = td }
 in
  match_thms tmorph
 end


fun thymorph_cmd norm src trg classMap_rep tpMap_rep opMap_rep axMap_rep =
 (case (Sigmorph.sigmorph_cmd src trg classMap_rep tpMap_rep opMap_rep) of
   Sigmorph.Just sigma => thymorph_by_sigmorph_cmd norm sigma axMap_rep
 | Sigmorph.AmbiguousConsts (cs, mp, sortmap, typemap) => 
   (let val _ = Aux.info (AWE.show_der_map mp)
        val _ = Aux.error_msg ("There are following ambiguous mappings:\n" ^ (AWE.show_amb_map cs))
        val _ = Aux.error_msg ("Please use the operation map to resolve.")
        val _ = Aux.error_msg ("Suggestion:")
        val _ = Aux.error_msg (AWE.plain_repr (AWE.mk_suggestion cs))                   
    in
     raise THYMORPH "Theory morphism construction failed."
    end))


(* **************************** *)
(* *** Extension        ******* *)
(* **************************** *)


fun cod_ext (ThyMorph {sigma, axiomMap, theoremMap, data=ThyMorphData td}) 
            thmMap_ext ntrg =
 let val o_trg = Sigmorph.trg sigma
     val nsigma = Sigmorph.cod_ext sigma ntrg
     val Sigmorph.SigMorphData d = Sigmorph.data nsigma
     val n1 = Aux.name_of_thy o_trg
     val n2 = Aux.name_of_thy ntrg
 in 
   if (n1 = n2) then 
    (let val new_thms = (let val thms1 = map fst (Symtab.dest (#cod_thms td)) 
                             val thms2 = (Aux.flat_thms_of ntrg)
                         in
                          filter (fn (nm, t) => (not (nm mem_string thms1))) thms2
                         end) 
         val new_thms' = prepare_cod_thms Symtab.empty new_thms
         val ndata = ThyMorphData {dom_thms = (#dom_thms td),
                                   cod_thms = Aux.extend_symtab (#cod_thms td, Symtab.dest new_thms') }
                     handle DUPS => error ("DUPS1 in Thymorph.cod_ext")
         val tmorph = ThyMorph {sigma = nsigma, axiomMap = axiomMap, 
                                theoremMap = (Symtab.merge (op =) (theoremMap, thmMap_ext)), 
                                data = ndata}
                      handle DUPS => error ("DUPS2 in Thymorph.cod_ext")   
         val tmorph' = (case new_thms of
                         [] => tmorph
                       | _ => match_thms tmorph)
     in
      tmorph'
     end)
   else (case (Aux.unq_path o_trg ntrg) of
          [] => thymorph_by_sigmorph_cmd false nsigma (transf_axMap (Symtab.dest axiomMap))
        | path => 
          if ((Aux.name_of_thy (Sigmorph.src nsigma)) mem_string (map Aux.name_of_thy path))
          then thymorph_by_sigmorph_cmd false nsigma (transf_axMap (Symtab.dest axiomMap))  
          else    (let val new_thms = Aux.map_foldl Aux.flat_thms_of op@ ([], path)  
                       val new_thms' =  prepare_cod_thms Symtab.empty new_thms
                       val ndata = ThyMorphData {dom_thms = (#dom_thms td),
                                                 cod_thms = Aux.extend_symtab (#cod_thms td, 
                                                                                  Symtab.dest new_thms') }
                                   handle DUPS => error ("DUPS3 in Thymorph.cod_ext")
                       val tmorph = ThyMorph {sigma = nsigma, axiomMap = axiomMap, 
                                              theoremMap = (Symtab.merge (op =) (theoremMap, thmMap_ext)), 
                                              data = ndata}
                                    handle DUPS => error ("DUPS4 in Thymorph.cod_ext")                                  
                       val tmorph' = (case new_thms of
                                       [] => tmorph
                                     | _ => match_thms tmorph)
                   in
                    tmorph' 
                   end))
 end


(* ********************* *)
(* *** Normalization *** *)
(* ********************* *)

fun normalize renames (ThyMorph {sigma, axiomMap, theoremMap, data}) =
 let val nsigma = Sigmorph.normalize renames sigma
 in
  ThyMorph{sigma = nsigma, axiomMap=axiomMap, 
           theoremMap=theoremMap, data = data}
 end

fun translate_syntax renames (ThyMorph {sigma, axiomMap, theoremMap, data}) =
 let val nsigma = Sigmorph.translate_syntax renames sigma
 in
  ThyMorph{sigma = nsigma, axiomMap=axiomMap, 
           theoremMap=theoremMap, data = data}
 end



(* **************************** *)
(* *** Theorem translations *** *)
(* **************************** *)



fun new_name names pre id = (if (id mem_string names) then new_name names pre (pre ^ "_" ^ id)
                             else id)


fun translate_thms sat src thy trg_id trg_names pre dom_thms sigma axiomMap theoremMap renames [] = 
    (thy, theoremMap)

  | translate_thms sat src thy trg_id trg_names pre dom_thms sigma axiomMap theoremMap renames (x::xs) =
(if (Symtab.defined theoremMap x) 
 then translate_thms sat src thy trg_id trg_names pre dom_thms sigma axiomMap theoremMap renames xs
 else (case (Symtab.lookup dom_thms x) of
        NONE => translate_thms sat src thy trg_id trg_names pre dom_thms sigma axiomMap theoremMap renames xs
      | SOME (thm, term, deps) =>
         (if ((deps = []) orelse (x mem_string sat)) then 
             (let val newname_ = (case (Symtab.lookup renames x) of
                                    NONE => new_name trg_names pre (snd (Aux.qname_split2 x))
                                  | SOME (r, _) => r)

                  val (path, newname) = Aux.qname_split newname_
                  val newname' = if path = "" then (trg_id ^ "." ^ newname)
                                 else (trg_id ^ "." ^ path ^ "." ^ newname)

                  val _ = Aux.info ("... translating theorem "^ x ^" as " ^ newname' ^ " ...") 

                  val newthm = (if !AWE.skip_proofs 
                                then (let val t = Aux.unvarify (prop_of thm)
                                          val mt = Sigmorph.map_term sigma t
                                          val _ = Aux.warning "proof skipped"
                                          val _ = (quick_and_dirty := true)
                                          val thm = (Drule.standard o Thm.varifyT) (Skip_Proof.make_thm thy mt) 
                                          val _ = (quick_and_dirty := false)
                                      in
                                       thm
                                     end)
                                else (let val prf = Aux.fullprf_of thm
                                          val defMap = Sigmorph.defMap sigma 
                                      in 
                                       Derive_Thm.derive_thm (Sigmorph.maps_class sigma)
                                                             (Sigmorph.map_type sigma) 
                                                             (Sigmorph.map_term sigma)
                                                             defMap axiomMap theoremMap src thy prf
                                       handle Derive_Thm.DERIVE_THM s =>
                                        error ("Theorem derivation for "^x^" failed: " ^ s)
                                      end))

                  val theoremMap' = Symtab.update_new (x, newname') theoremMap
                                     handle DUP => error "Unexpected DUP in translate_thm_i"
                  val (gen, thy') = 
                  (if (path = "") then PureThy.store_thm (Binding.name newname, newthm) thy
                   else thy 
                        |> Sign.add_path path
                        |> PureThy.store_thm (Binding.name newname, newthm)
                        ||> Sign.parent_path)
                  
                  val _ = Aux.info ("... added theorem " ^ newname' ^ " : \n" ^ 
                                    (Display.string_of_thm (ProofContext.init thy') newthm))
              in
               translate_thms sat src thy' trg_id trg_names pre dom_thms sigma axiomMap theoremMap' renames xs
              end)
          else translate_thms (x::sat) src thy trg_id 
               trg_names pre dom_thms sigma axiomMap theoremMap renames (deps @ (x::xs)))))  



fun translate_thm_i (tmorph as (ThyMorph {sigma, axiomMap, theoremMap, data = ThyMorphData td, ...})) 
                     renames id =
(case (Symtab.lookup theoremMap id) of
  SOME i => (let val _ = Aux.warning ("Theorem " ^ id ^ " is already mapped to " ^ i)
             in
               (tmorph, SOME (i, Aux.get_thm (Sigmorph.trg sigma) i))
             end)
| NONE => (case (Symtab.lookup (#dom_thms td) id) of
             NONE => (let val thm = SOME (Aux.get_thm (Sigmorph.trg sigma) id)
                                    handle e => (let val _ = writeln ("Identifier " ^ id ^ 
                                                                      " is neither local nor global")
                                                 in
                                                  NONE
                                                 end)
                      in
                        case thm of
                          SOME t => (tmorph, SOME(id, t))
                        | NONE => (tmorph, NONE)
                      end)
           | SOME (thm, term, dep) => 
              (let val trg = Sigmorph.trg sigma
                   val trg_id = Aux.name_of_thy trg
                   val trg_names = map (snd o Aux.qname_split2 o fst) (Aux.thms_of trg)
                   val src = Sigmorph.src sigma
                   val src_id = Aux.name_of_thy src
                   val Sigmorph.SigMorphData d = Sigmorph.data sigma

                   val newname_ = (case (Symtab.lookup renames id) of
                                    NONE => new_name trg_names src_id (snd (Aux.qname_split2 id))
                                  | SOME (r, _) => r)
                   val (path, newname) = Aux.qname_split newname_
                   val newname' = if path = "" then (trg_id ^ "." ^ newname)
                                  else (trg_id ^ "." ^ path ^ "." ^ newname)

                   val _ = Aux.info ("Translating theorem "^ id ^" as " ^ newname' ^ " ...") 

                   val (trg', theoremMap') = translate_thms [] src trg trg_id trg_names src_id 
                                             (#dom_thms td) sigma axiomMap theoremMap renames dep
                   val newthm = 
                    (case ((try (Goal.prove_global trg' [] [] (Sigmorph.map_term sigma term))) 
                           (fn x => Class.intro_classes_tac (#prems x))) of
                      SOME thm_ => (Drule.standard o Thm.varifyT o Aux.beta_eta_convert) thm_
                    | NONE =>(if !AWE.skip_proofs 
                                 then (let val t = Aux.unvarify (prop_of thm)
                                           val mt = Sigmorph.map_term sigma t
                                           val _ = Aux.warning "proof skipped"
                                           val _ = (quick_and_dirty := true)
                                           val thm = (Drule.standard o Thm.varifyT) (Skip_Proof.make_thm trg' mt) 
                                           val _ = (quick_and_dirty := false)
                                       in
                                        thm
                                      end)
                                else (let val prf = Aux.fullprf_of thm  
                                          val defMap = Sigmorph.defMap sigma 
                                      in 
                                       Derive_Thm.derive_thm (Sigmorph.maps_class sigma)
                                                             (Sigmorph.map_type sigma) 
                                                             (Sigmorph.map_term sigma) 
                                                             defMap axiomMap theoremMap' src trg' prf
                                        handle Derive_Thm.DERIVE_THM s =>
                                        error ("Theorem derivation for "^id^" failed: " ^ s)
                                      end)))

                   val (gen, trg'') = 
                  (if (path = "") then PureThy.store_thm (Binding.name newname, newthm) trg'
                   else trg' 
                        |> Sign.add_path path
                        |> PureThy.store_thm (Binding.name newname, newthm)
                        ||> Sign.parent_path)                  
                   val theoremMap'' = Symtab.update_new (id, newname') theoremMap'
                                      handle DUP => error "Unexpected DUP in translate_thm_i"
                   val _ = Aux.info ("... added theorem " ^ newname' ^ " :\n" ^ 
                                    (Display.string_of_thm (ProofContext.init trg'') newthm))
                   val tmorph' = cod_ext tmorph theoremMap'' trg''
               in
                (tmorph', SOME (newname', newthm))
               end)))


fun translate_thm' tmorph' renames id =
 let val sigma = sigmorph tmorph'
     val src = Sigmorph.src sigma
     val thms = PureThy.get_thms src id
     fun get_ids [] = []
       | get_ids (t::ts) = (case (Aux.name_of_thm t) of
                              SOME x => x::(get_ids ts)
                            | NONE => (let val _ = Aux.warning ("No tag for name found")
                                       in
                                         get_ids ts
                                       end))    
 in
  (case thms of
     [] => error "no thms"
   | [x] => (let val (tmorph'', opt) = translate_thm_i tmorph' renames id
             in
              case opt of
               NONE => (tmorph'', [])
             | SOME r => (tmorph'', [r])
             end)
   
   | xs => 
     (let val ids = map Thm.get_name xs
          fun translate ((mrph, translated), x) = 
             (let val (mrph', opt) = translate_thm_i mrph renames x
              in
               case opt of
                NONE => error "Unexpected error: empty result in translate_thm"
              | SOME r => (mrph', translated@[r])
              end)       
          val (tmorph'', tthms) = Aux.foldl translate ((tmorph', []), ids) 

          val trg = Sigmorph.trg (sigmorph tmorph'')
          val trg_id = Aux.name_of_thy trg
          val trg_names = map (snd o Aux.qname_split2 o fst) (Aux.thms_of trg)
          val src = Sigmorph.src (sigmorph tmorph'')
          val src_id = Aux.name_of_thy src

          val newname_ = (case (Symtab.lookup renames id) of
                           NONE => new_name trg_names src_id (snd (Aux.qname_split2 id))
                         | SOME (r, _) => r)
          val (path, newname) = Aux.qname_split newname_
          val newname' = if path = "" then (trg_id ^ "." ^ newname)
                         else (trg_id ^ "." ^ path ^ "." ^ newname)
          fun note nm thms thy =
           (case (try (PureThy.note_thmss "lemmas" 
                        [(Thm.no_attributes (Binding.name nm), 
                        map (fn (n, t) => ([t], [])) thms)]) thy) of
              NONE => note (nm ^ "'") thms thy
            | SOME x => x)

          val (gen, trg') = 
          (if (path = "") then PureThy.note_thmss "lemmas" 
                               [(Thm.no_attributes (Binding.name newname), 
                                 map (fn (n, t) => ([t], [])) tthms)] trg
           else trg
                |> Sign.add_path path
                |> note newname tthms
                ||> Sign.parent_path)

          val tmorph''' = cod_ext tmorph'' Symtab.empty trg'
          val _ = Aux.info ("Added lemmas " ^ newname')
     in
       (tmorph''', tthms)
     end))
 end

fun translate_thm tmorph renames id =
 let val tmorph' = normalize renames tmorph
 in
  translate_thm' (match_thms tmorph') renames id
 end


(* **************************** *)
(* *** Composition      ******* *)
(* **************************** *)

(* compose (tm1 : T1 --> T2) (tm2 : T3 --> T4) =def= (tm3 : T1 --> T4, tm2_norm) 
   where T2 in Anc(T3) (tm2_norm is normalized tm2). *)


fun compose renames
            (tm1 as (ThyMorph {sigma=sigma1, axiomMap=axiomMap1, theoremMap=theoremMap1, ...}))
            (tm2 as (ThyMorph {sigma=sigma2, axiomMap=axiomMap2, theoremMap=theoremMap2, data=data2, ...})) =
 let val (sc, nsigma2) = Sigmorph.compose renames sigma1 sigma2
     val ntm2 = ThyMorph{sigma = nsigma2, axiomMap=axiomMap2, 
                         theoremMap=theoremMap2, data = data2}
     val ntm2' = match_thms ntm2

     val Sigmorph.SigMorphData d = Sigmorph.data sc
     val basis = Symtab.dest(#dom_axs_map d)

     fun compMaps tm2u a [] = (tm2u, a)
       | compMaps tm2u a ((nm, _)::bs) = 
     (case (Symtab.lookup axiomMap1 nm) of
      SOME (Derive_Thm.Axm i) => 
                 (case (Symtab.lookup axiomMap2 i) of
                         SOME (Derive_Thm.Axm i') => compMaps tm2u 
                                                     (Symtab.update_new (nm, Derive_Thm.Axm i') a) bs
                       | SOME (Derive_Thm.Thm i') => compMaps tm2u 
                                                     (Symtab.update_new (nm, Derive_Thm.Thm i') a) bs
                       | NONE =>  compMaps tm2u (Symtab.update_new (nm, Derive_Thm.Axm i) a) bs)
    | SOME (Derive_Thm.Thm i) => 
                 (case (Symtab.lookup (theoremMap tm2u) i) of
                        SOME i' => compMaps tm2u (Symtab.update_new (nm, Derive_Thm.Thm i') a) bs
                      | NONE => (let val (tm2u', i') = (case (translate_thm' tm2u renames i) of
                                                         (t, [(x, _)]) => (t, x)
                                                       | (t, _) => error ("Unexpected result of theorem\
                                                                              \ translation in compose")) 
                                 in
                                  compMaps tm2u' (Symtab.update_new (nm, Derive_Thm.Thm i') a) bs
                                 end))
    | NONE => (case (Symtab.lookup axiomMap2 nm) of
                SOME (Derive_Thm.Axm i') => compMaps tm2u (Symtab.update_new (nm, Derive_Thm.Axm i') a) bs
              | SOME (Derive_Thm.Thm i') => compMaps tm2u (Symtab.update_new (nm, Derive_Thm.Thm i') a) bs
              | NONE => error "Unexpected global axiom in compose: mk_axMaps"))

     val (tm2u, ax_map) = compMaps ntm2' Symtab.empty basis
     val ntrg = Sigmorph.trg(sigmorph tm2u)
     val sc' = Sigmorph.cod_ext sc ntrg
     val td = get_thymorph_data sc'
     val theoremMap2 = theoremMap tm2u

     fun mk_thmMap m [] = m
       | mk_thmMap m ((a, b)::xs) = 
      (case (Symtab.lookup theoremMap2 b) of
        SOME i => (let val _ = (if (Aux.internal_id a) then ()
                                else Aux.info ("Composed: (" ^ (quote a)
                                               ^ " |-> " ^ (quote i) ^ ")"))
                   in
                    mk_thmMap (Symtab.update (a, i) m) xs
                   end)
      | NONE => mk_thmMap m xs)

      val ctmorph = ThyMorph{sigma=sc', axiomMap=ax_map, theoremMap=Symtab.empty, data = td}
 in
  (match_thms ctmorph, tm2u)
 end
 



(* **************************** *)
(* *** Instantiation    ******* *)
(* **************************** *)

datatype Inst_Result = ThyM of thymorph 
                     | SigM of (Sigmorph.sigmorph * ((string * Thm.cterm) list))




 


fun instantiate axiomatic syn class_specs type_specs op_specs axm_specs renames thy 
     (tmorph as (ThyMorph {sigma, axiomMap=axiomMap_, theoremMap, ...})) =
let val (nsigma, axiomMap') = Sigmorph.instantiate syn class_specs type_specs op_specs renames thy sigma
    val Sigmorph.SigMorphData d = Sigmorph.data nsigma
    val (ThyMorphData td) = get_thymorph_data nsigma
    val axs_basis = Symtab.dest(#dom_axs_map d)
    val cod_thms = Symtab.dest(#cod_thms td)

    val axiomMap'' = ext_mapping axiomMap_ (Symtab.dest axiomMap')
    val (nsigma', axiomMap) = match_axms false nsigma axiomMap'' cod_thms

    val axs_basis = Symtab.dest(#dom_axs_map d)
    val src = Sigmorph.src nsigma'
    val trg = Sigmorph.trg nsigma'
    val trg_id = Aux.name_of_thy trg

    fun check_axm_specs trg src_term nm =
     (case (Aux.assoc_string (axm_specs, nm)) of
       NONE => NONE
     | SOME tnm => (let val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
                        val (trg_term, b) = (norm (Aux.get_axiom trg tnm), true)
                                             handle e => (norm (Aux.get_thm trg tnm), false)
                    in
                      if (Alpha.term_aconv (Sign.tsig_of trg) (src_term, trg_term))
                      then SOME (tnm, b)
                      else (let val _ = Aux.warning ("Incorrect explicit axiom instantiation: " ^
                                        nm ^ " |-> " ^ tnm)
                            in
                             NONE
                            end)
                    end)) 
    fun goals sigma axmap [] = []
      | goals sigma axmap ((nm, ax)::xs) =
      (case (Symtab.lookup axmap nm) of
        NONE => (let val term = Sigmorph.map_term sigma ax
                     val cterm = Thm.cterm_of trg term
                     val nnm = (case (Symtab.lookup renames nm) of
                                 SOME (r, _) => r
                               | NONE => Aux.base nm)
                 in
                  (nnm, cterm)::(goals sigma axmap xs)
                 end)
      | _ => goals sigma axmap xs)

    fun ins_axs_basis a trg [] = (a, trg)
      | ins_axs_basis a trg ((nm, term)::bs) =
    let val tterm = (Sigmorph.map_term nsigma' term)
    in
      (case (Symtab.lookup a nm) of
              SOME i => ins_axs_basis a trg bs
            | NONE => 
            (case (check_axm_specs trg tterm nm) of
              NONE => (let val (nnm, trg') = Transfer.transfer_axiom (thy, nm) (Sigmorph.map_term nsigma')
                                             renames I trg
                           val a' = Symtab.update_new (nm, Derive_Thm.Axm nnm) a
                       in
                        ins_axs_basis a' trg' bs
                       end)
           | SOME (nnm, b) => 
              (if b then (let val a' = Symtab.update_new (nm, Derive_Thm.Axm nnm) a
                              val _ = Aux.info ("... axiom " ^ nm ^ " explicitly instantiated")
                          in
                           ins_axs_basis a' trg bs
                          end)                      
               else (let val a' = Symtab.update_new (nm, Derive_Thm.Thm nnm) a
                         val _ = Aux.info ("... axiom " ^ nm ^ " explicitly instantiated")
                     in
                      ins_axs_basis a' trg bs
                     end)))) 
     end   

in 
 if (not axiomatic) 
 then (let val gs = goals nsigma' axiomMap axs_basis
       in
        case gs of
         [] => (let val ntmorph = ThyMorph { sigma = nsigma', axiomMap = axiomMap, 
                                             theoremMap = theoremMap, data = ThyMorphData td}
                in
                 ThyM (match_thms ntmorph)
                end)
       | _ => SigM (nsigma', gs)
       end)
 else (let val (axiomMap', trg') = ins_axs_basis axiomMap trg axs_basis
           val nsigma'' = Sigmorph.cod_ext nsigma' trg'
           val td' = get_thymorph_data nsigma''
           val ntmorph = ThyMorph { sigma = nsigma'', axiomMap = axiomMap', 
                                    theoremMap = theoremMap, data = td' }
      in
       ThyM (match_thms ntmorph)
      end)
end



(* **************************** *)
(* *** Thymorph goals   ******* *)
(* **************************** *)

fun thymorph_goals syn renames isigma =
 let val sigma = (if syn then Sigmorph.translate_syntax renames isigma
                  else Sigmorph.normalize renames isigma)
     val src = Sigmorph.src sigma
     val trg = Sigmorph.trg sigma
     val (Sigmorph.SigMorphData d) = Sigmorph.data sigma
     val (ThyMorphData td) = get_thymorph_data sigma

     val axs_basis = Symtab.dest(#dom_axs_map d)
     val cod_thms = Symtab.dest(#cod_thms td)

     val (sigma', axmMap) = match_axms false sigma Symtab.empty cod_thms  
     val trg' = Sigmorph.trg sigma'    
 
     fun mk_goals [] = []
       | mk_goals ((nm, ax)::xs) =
      (case (Symtab.lookup axmMap nm) of 
        NONE => (let val term = Sigmorph.map_term sigma' ax
                     val cterm = Thm.cterm_of trg' term
                     val nnm = (case (Symtab.lookup renames nm) of
                                 SOME (r, _) => r
                               | NONE => Aux.base nm)
                 in
                  (nnm, cterm)::(mk_goals xs)
                 end)
      | _ => mk_goals xs)

     val goals = mk_goals axs_basis
 in
  (sigma', goals)
 end


(* ***************************** *)
(* *** String representation *** *)
(* ***************************** *)


fun thymorph_to_string (ThyMorph {sigma, axiomMap, ...}) =
 let fun repr (a, b) = (case b of
                         Derive_Thm.Axm i => ("(" ^ (quote a) ^ " |-> Axm " ^ (quote i) ^ ")")
                       | Derive_Thm.Thm i => ("(" ^ (quote a) ^ " |-> Thm " ^ (quote i) ^ ")")) 
     fun sep [] = ""
       | sep [s] = s
       | sep (x::xs) = x^",\n "^(sep xs)
     val sigma_s = Sigmorph.sigmorph_to_string sigma
     val axiomMap_s = ("\naxiom map:\n{" ^ (sep (map repr (Symtab.dest axiomMap))) ^ "}\n")
 in 
  sigma_s ^ axiomMap_s
 end


(* **************************** *)
(* *** Managing thymorphs ***** *)
(* **************************** *)

val thymorphs = Unsynchronized.ref (Symtab.empty : thymorph Symtab.table)

fun add_thymorph x = (thymorphs := (Symtab.update x (!thymorphs)))
fun get_thymorph id = (Symtab.lookup (!thymorphs) id)
fun defined_thymorph id = (case (get_thymorph id) of
                            NONE => false
                          | _ => true)
fun add_thymorph_verb (id, tmorph) = 
 (let val _ = (if (defined_thymorph id)
               then Aux.info ("### Theory morphism " ^ id ^ " already exists,\
                                                  \ overwriting. ")
               else ())
  in
   add_thymorph (id, tmorph)
  end)
fun delete_thymorph id = (thymorphs := (Symtab.delete id (!thymorphs)))

fun print_thymorphs () =
   let fun prnt (s, (nm, tmorph)) = let val srcnm = Aux.name_of_thy (Sigmorph.src (sigmorph tmorph))
                                        val trgnm = Aux.name_of_thy (Sigmorph.trg (sigmorph tmorph))
                                   in
                                              (s ^  (nm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n"))
                                   end
   in
    Aux.foldl prnt ("", Symtab.dest(!thymorphs)) 
   end    







(* **************************** *)
(* *** Parsers          ******* *)
(* **************************** *)


(* ------------------------------------------------------------------------- *)
(* set up the 'thymorph' command                                               *)
(* ------------------------------------------------------------------------- *)


val scan_col  = OuterParse.$$$ ":"
val scan_arrow = ((OuterParse.$$$ "-->") || (OuterParse.$$$ "\<longrightarrow>"))
val scan_classmp = Scan.option ((OuterParse.$$$ "class_map" -- scan_col) |-- Aux.scan_mapping)
val scan_typemp = Scan.option ((OuterParse.$$$ "type_map" -- scan_col) |-- Aux.scan_mapping)
val scan_opmp = Scan.option ((OuterParse.$$$ "op_map" -- scan_col) |-- Aux.scan_mapping)
val scan_axmp = Scan.option ((OuterParse.$$$ "axiom_map" -- scan_col) |-- Aux.scan_mapping)
val scan_mrphnm =OuterParse.name 
val scan_srcnm = OuterParse.name
val scan_trgnm = OuterParse.name
val scan_by_sigmorph = OuterParse.$$$ "by_sigmorph"
val scan_by_comp = OuterParse.$$$ "by_composition"
val scan_circ  = ((OuterParse.$$$ "o") || (OuterParse.$$$ "\<circ>"))
val scan_ren_key = OuterParse.$$$ "renames" -- OuterParse.$$$ ":"
val scan_renames = Scan.option(scan_ren_key |-- Aux.scan_renaming)

fun lift1 f t =
    let val ((srcnm, (trgnm, (opt1, (opt2, (opt3, opt4))))), t') = f t
    in 
     ((SOME(srcnm, trgnm, opt1, opt2, opt3, opt4), NONE, NONE), t')
    end


fun lift2 f t =
    let val ((smrphnm, opt), t') = f t
    in 
      ((NONE, SOME (smrphnm, opt), NONE), t')
    end

fun lift3 f t =
    let val ((m2, (m1, ren)), t') = f t
    in
     ((NONE, NONE, SOME (m1, m2, ren)), t')
    end


val thymorph_scan_args = 
 (scan_mrphnm --
  ((lift1 (scan_col |-- (scan_srcnm -- (scan_arrow |-- 
           (scan_trgnm -- (scan_classmp -- (scan_typemp -- (scan_opmp -- scan_axmp))))))))
   ||
   (lift2 (scan_by_sigmorph |-- (scan_mrphnm -- scan_axmp)))
   ||
   (lift3 (scan_by_comp |-- (scan_mrphnm -- (scan_circ |-- (scan_mrphnm -- scan_renames)))))) )




fun flat_map [] = []
  | flat_map ((ls, x)::mp) =
    let fun flat b [] = []
          | flat b (a::xs) = (a, b)::(flat b xs)
    in
     (flat x ls)@(flat_map mp)
    end



fun thymorph_trans args =
Toplevel.theory 
( fn thy => 
 (let val (mrphnm, (opt1, opt2, opt3)) = args
  in
  (case opt1 of
    SOME (srcnm, trgnm, opt_class, opt_tpmp, opt_opmp, opt_axmp) =>
      (let val src = if (Aux.name_of_thy thy = srcnm) then thy else (theory srcnm)
                                      handle e => (let val _ = ThyInfo.use_thy srcnm
                                                   in (theory (Aux.path_base srcnm)) end)

           val trg = if (Aux.name_of_thy thy = trgnm) then thy else (theory trgnm)
                                      handle e => (let val _ = ThyInfo.use_thy trgnm
                                                   in (theory (Aux.path_base trgnm)) end)

           fun res NONE = []
             | res (SOME x) = flat_map x

           val tmorph = thymorph_cmd false src trg (res opt_class) (res opt_tpmp) 
                                                   (res opt_opmp) (res opt_axmp)
           val _ = Aux.info ("Theory morphism " ^ mrphnm ^ 
                             " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.")
           val _ = add_thymorph_verb (mrphnm, tmorph)
           val ntrg = Sigmorph.trg (sigmorph tmorph)
           val ntrg_id = Aux.name_of_thy ntrg 
       in
         if (ntrg_id = (Aux.name_of_thy thy)) then ntrg else thy
       end)
  | NONE =>
    (case opt2 of
      SOME (sigma_id, opt_axmp) => 
       (let val sigma = (case (Sigmorph.get_sigmorph sigma_id) of
                           NONE => error ("Unknown signature morphism: " ^ sigma_id)
                         | SOME x => x)
            val src = Sigmorph.src sigma
            val trg = Sigmorph.trg sigma
            val srcnm = Aux.name_of_thy src
            val trgnm = Aux.name_of_thy trg

            val sigma' = (if (Aux.name_of_thy thy = trgnm)
                          then Sigmorph.cod_ext sigma thy
                          else Sigmorph.cod_ext sigma (theory trgnm))
            val axMap_rep = (case opt_axmp of
                               NONE => []
                             | SOME x => flat_map x)
            val tmorph = thymorph_by_sigmorph_cmd false sigma' axMap_rep
            val _ = Aux.info ("Theory morphism " ^ mrphnm ^ 
                              " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.")
            val _ = add_thymorph_verb (mrphnm, tmorph)
            val ntrg = Sigmorph.trg (sigmorph tmorph)
            val ntrg_id = Aux.name_of_thy ntrg 
        in
          if (ntrg_id = (Aux.name_of_thy thy)) then ntrg else thy
        end)
    | NONE =>
     (let val SOME (tnm1, tnm2, ren) = opt3
              handle e => error "Unexpected parsing result in thymorph_trans"

          val t1 = case (get_thymorph tnm1) of
                      SOME t => t
                    | NONE   => error ("Unknown theory morphism " ^ tnm1)
          val t2 = case (get_thymorph tnm2) of
                      SOME t => t
                    | NONE   => error ("Unknown theory morphism " ^ tnm2)
   
          val renames = case ren of 
                          NONE => Symtab.empty
                        | SOME r => Aux.foldr (uncurry Symtab.update) (r, Symtab.empty)

          val sigma1 = sigmorph t1
          val sigma2 = sigmorph t2
          val source = Sigmorph.src sigma1
          val srcnm  = Aux.name_of_thy source
          val trg1   = Sigmorph.trg sigma1
          val trg1nm = Aux.name_of_thy trg1
          val src2   = Sigmorph.src sigma2
          val src2nm = Aux.name_of_thy src2  
          val target = Sigmorph.trg sigma2
          val trgnm  = Aux.name_of_thy target
          val thisnm  = Aux.name_of_thy thy

          val t2' = (if (thisnm = trgnm) then cod_ext t2 Symtab.empty thy 
                     else (if (trgnm mem_string (map Aux.name_of_thy (Theory.ancestors_of thy)))
                           then cod_ext t2 Symtab.empty thy
                           else cod_ext t2 Symtab.empty (theory trgnm))) 
        
          val _ = Aux.info ("Composing theory morphisms:\n" ^
                             "(" ^ tnm2 ^ " : " ^ src2nm ^ " --> " ^ trgnm ^ ") o (" 
                                 ^ tnm1 ^ " : " ^ srcnm ^ " --> " ^ trg1nm ^ ") ...")
                               
          val (tc, t2u) = compose renames t1 t2'
          val nthy   = Sigmorph.trg (sigmorph tc)
   
          val _ = Aux.info ("Theory morphism " ^ mrphnm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.") 
          val _ = add_thymorph_verb (mrphnm, tc)
          val _ = add_thymorph (tnm2, t2u)
      in
       if (Aux.name_of_thy nthy = thisnm) then nthy else thy
      end)))
  end))



fun thymorph_parser tokens = (thymorph_scan_args tokens) |>> thymorph_trans 
                                        
val _ = OuterSyntax.command "thymorph" 
                   "Construct a theory morphism between two theories by given class, type, operation\
                   \ and axiom maps, by given signature morphism and axiom map or as composition \
                   \ of existing theory morphisms" 
                   OuterKeyword.thy_decl thymorph_parser;



(* ------------------------------------------------------------------------- *)
(* set up the 'translate_thm' command                                        *)
(* ------------------------------------------------------------------------- *)

val scan_tmrphnm = OuterParse.name; 
val scan_thmnm = OuterParse.string;
val scan_along  = OuterParse.$$$ "along";
val scan_rename = Scan.option(OuterParse.$$$ "as" |-- OuterParse.name);

val translate_thm_scan = scan_thmnm -- (scan_rename -- (scan_along |-- (scan_tmrphnm -- scan_renames)))

fun translate_thm_trans args =
  Toplevel.theory (fn thy =>
  ( let val (thmnm, (opt_thmrnm, (tmrphnm, ren))) = args
        val renames_ = case ren of
                        NONE => Symtab.empty
                      | SOME r => Aux.foldr (uncurry Symtab.update) (r, Symtab.empty)
        val renames = (case opt_thmrnm of
                        NONE => renames_
                      | SOME x => Symtab.update (thmnm, (x, NONE)) renames_) 
        val tm = case (get_thymorph tmrphnm) of
                    SOME t => t
                  | NONE   => error ("Unknown theory morphism " ^ tmrphnm)
  
        val sigma = sigmorph tm
        val src   = Sigmorph.src sigma
        val srcnm = Aux.name_of_thy src
        val trg   = Sigmorph.trg sigma
        val trgnm = Aux.name_of_thy trg
        val thisnm  = Aux.name_of_thy thy

        val tm' = (if (thisnm = trgnm) then cod_ext tm Symtab.empty thy 
                     else (if (trgnm mem_string (map Aux.name_of_thy (Theory.ancestors_of thy)))
                           then cod_ext tm Symtab.empty thy
                           else cod_ext tm Symtab.empty (theory trgnm))) 

        val (tm'', _) = translate_thm tm' renames thmnm
        val _ = add_thymorph(tmrphnm, tm'')

        val ntrg = Sigmorph.trg(sigmorph tm'')
        val newthy = (if (thisnm = Aux.name_of_thy ntrg) then ntrg
                      else thy)
    in
     newthy
    end))     



fun translate_thm_parser tokens = translate_thm_scan tokens |>> translate_thm_trans

val _ = OuterSyntax.command "translate_thm" "Translate a theorem along a theory morphism" 
                         OuterKeyword.thy_decl translate_thm_parser;




(* ------------------------------------------------------------------------- *)
(* set up the 'instantiate_theory' command                                   *)
(* ------------------------------------------------------------------------- *)

val scan_class_specs = scan_classmp
val scan_type_specs = scan_typemp
val scan_op_specs = scan_opmp
val scan_axm_specs = scan_axmp
val scan_syn_key = Scan.option(OuterParse.$$$ "without_syntax")
val scan_ax = Scan.option(OuterParse.$$$ "axiomatic")


val instantiate_scan = OuterParse.name -- (OuterParse.$$$ "by_thymorph" |-- (OuterParse.name -- 
                       (scan_class_specs -- (scan_type_specs -- (scan_op_specs -- (scan_axm_specs -- 
                       (scan_renames -- (scan_ax -- scan_syn_key))))))))

fun inst_trans args =
     Toplevel.theory (fn thy => 
 (let val (intrgnm, (mnm, (cl_specs_, (type_specs_, (op_specs_, (axm_specs_, (ren, (ax_opt, syn)))))))) = args
      val thynm = Aux.name_of_thy thy

      val intrg = (if (thynm = intrgnm) then thy 
                   else (theory intrgnm)
                        handle e => (let val _ = ThyInfo.use_thy intrgnm
                                     in (theory (Aux.path_base intrgnm)) end))

                                     
      val tmorph = (case (get_thymorph mnm) of
                     NONE => error ("Unknown theory morphism " ^ mnm)
                   | SOME m => m)
      val src = Sigmorph.src (sigmorph tmorph) 
      val trg = Sigmorph.trg (sigmorph tmorph)
      val trgnm = Aux.name_of_thy trg
                                     
      val tmorph' = (if (trgnm = thynm) then cod_ext tmorph Symtab.empty thy 
                     else (if (trgnm mem_string (map Aux.name_of_thy (Theory.ancestors_of thy)))
                           then cod_ext tmorph Symtab.empty thy
                           else cod_ext tmorph Symtab.empty (theory trgnm)))                 
      val class_specs = case cl_specs_ of
                          NONE => []
                        | SOME xs => flat_map xs  
      val type_specs = case type_specs_ of
                        NONE => []
                      | SOME xs => flat_map xs
      val op_specs = case op_specs_ of
                        NONE => []
                      | SOME xs => flat_map xs    
      val renames = case ren of
                      NONE => Symtab.empty
                    | SOME r => Aux.foldr (uncurry Symtab.update) (r, Symtab.empty)
      val axiomatic = case ax_opt of
                        NONE => false
                      | _ => true
      val axm_specs = case axm_specs_ of
                        NONE => []
                      | SOME xs => (if ((not axiomatic) andalso (xs <> [])) 
                                    then (let val _ = Aux.warning ("Non-axiomatic: explicit \
                                                      \axiom instantiation(s) will be ignored.")
                                          in
                                           flat_map xs
                                          end)
                                    else flat_map xs)

      val syntax = case syn of
                    NONE => true
                  | _ => false

      val _ = (if axiomatic then Aux.info ("Instantiating theory " ^ intrgnm ^ 
                                         " by theory morphism " ^ mnm ^ 
                                         " : " ^ (Aux.name_of_thy src) ^ " --> " ^ trgnm ^ " axiomatically ...")
               else  Aux.info ("Instantiating theory " ^ intrgnm ^ 
                                  " by theory morphism " ^ mnm ^ 
                                  " : " ^ (Aux.name_of_thy src) ^ " --> " ^ trgnm ^ " ..."))

      val result = instantiate axiomatic syntax class_specs type_specs op_specs axm_specs renames intrg tmorph'
      
  in
   case result of
    ThyM tmorph'' =>  (let val sigma'' = sigmorph tmorph''
                           val nsrc_nm = Aux.name_of_thy (Sigmorph.src sigma'')
                           val ntrg = Sigmorph.trg sigma''
                           val ntrg_nm = Aux.name_of_thy ntrg

                           val _ = Aux.info ("Theory morphism " ^ mnm ^ "' : " ^ 
                                                 nsrc_nm ^ " --> " ^ ntrg_nm ^ " constructed.") 
                           val _ = add_thymorph (mnm ^ "'", tmorph'')
                       in 
                         if (ntrg_nm = thynm) then ntrg else thy
                       end)
 | SigM (sigma'', goals) => (let val nsrc_nm = Aux.name_of_thy (Sigmorph.src sigma'')
                                 val ntrg = Sigmorph.trg sigma''
                                 val ntrg_nm = Aux.name_of_thy ntrg
                                
                                 fun show_goals [] = ()
                                   | show_goals ((nm, cterm)::xs) =
                                     (let val pp = (quote o (Syntax.string_of_term_global ntrg) o term_of) cterm
                                          val _ = Output.writeln ("lemma " ^ nm ^ ":\n" ^ 
                                                                 pp ^ "\n")
                                      in
                                       show_goals xs
                                      end)
                                    
                                 val _ = Aux.info ("Signature morphism " ^ mnm ^ "' : " ^ 
                                                       nsrc_nm ^ " --> " ^ ntrg_nm ^ " constructed.") 
                                 val _ = Sigmorph.add_sigmorph (mnm ^ "'", sigma'')
                                 val _ = Aux.info ("To prove:")
                                 val _ = show_goals goals 
                             in 
                                if (ntrg_nm = thynm) then ntrg else thy
                             end)
  end))


fun inst_parser tokens = instantiate_scan tokens |>> inst_trans

val _ = OuterSyntax.command "instantiate_theory" 
        "Instantiate a theory by a theory morphism" OuterKeyword.thy_decl inst_parser;


(* *** thymorph_goals *** *)

fun thymorph_goals_trans (morphnm, (ren, opt_syn)) =
       (Toplevel.theory 
          (fn thy => (let val thynm = Aux.name_of_thy thy
                          val morph = (case (Sigmorph.get_sigmorph morphnm) of
                                             SOME t => t
                                           | NONE => error ("Unknown signature morphism " ^ morphnm))
                          val renames = case ren of
                                          NONE => Symtab.empty
                                        | SOME r => Aux.foldr (uncurry Symtab.update) (r, Symtab.empty)
 
                          val target = Sigmorph.trg morph
                          val trgnm = Aux.name_of_thy target
                          val morph' = 
                          (if (trgnm = thynm) then Sigmorph.cod_ext morph thy 
                           else (if (trgnm mem_string (map Aux.name_of_thy (Theory.ancestors_of thy)))
                                 then Sigmorph.cod_ext morph thy
                                 else Sigmorph.cod_ext morph (theory trgnm))) 

                          val (morph'', goals) = (case opt_syn of
                                                    NONE => thymorph_goals true renames morph'
                                                  | _ => thymorph_goals false renames morph')
                          val ntrg = Sigmorph.trg morph''
                          val ntrg_nm = Aux.name_of_thy ntrg

                          fun show_goals [] = ()
                            | show_goals ((nm, cterm)::xs) =
                              (let val pp = (quote o (Syntax.string_of_term_global ntrg) o term_of) cterm
                                   val _ = Output.writeln ("lemma " ^ nm ^ ":\n" ^ 
                                                           pp ^ "\n")
                               in
                                show_goals xs
                               end)

                          val _ = Sigmorph.add_sigmorph (morphnm, morph'')
                          val _ = (case goals of
                                    [] => Aux.info ("No goals.")
                                   | _ => Aux.info ("To prove:"))
                          val _ = show_goals goals
                      in
                       if (ntrg_nm = thynm) then ntrg else thy
                      end)))

fun thymorph_goals_parser tokens = 
 (OuterParse.name -- (scan_renames -- Scan.option(OuterParse.$$$ "without_syntax"))) tokens
                                  |>> thymorph_goals_trans

val _ = OuterSyntax.command "thymorph_goals" 
                     "Show goals which have to be proven in order to extend a given signature morphism \
                     \ to a theory morphism" 
                     OuterKeyword.thy_decl thymorph_goals_parser;




(* Show morphisms *)

val _ = OuterSyntax.improper_command "print_sigmorphs" 
"Show the list of the registered signature morphisms" 
OuterKeyword.diag (fn x => 
                   (Toplevel.no_timing o (Toplevel.keep ( fn state =>
                                        (let val res = Sigmorph.print_sigmorphs ()
                                         in
                                            Aux.info res
                                         end))), x))

val _ = OuterSyntax.improper_command "print_thymorphs" 
"Show the list of the registered theory morphisms" 
OuterKeyword.diag (fn x => 
                   (Toplevel.no_timing o (Toplevel.keep ( fn state =>
                                        (let val res = print_thymorphs ()
                                         in
                                            Aux.info res
                                         end))), x))


fun print_sigmorph_trans arg =
Toplevel.keep ( fn state =>
                      (let val sigma = case (Sigmorph.get_sigmorph arg) of
                                        SOME s => s
                                      | NONE   => error ("Unknown signature morphism " ^ arg)

                           val srcnm  = Aux.name_of_thy (Sigmorph.src sigma)
                           val trgnm  = Aux.name_of_thy (Sigmorph.trg sigma)
                           val thynm  = Aux.name_of_thy (Toplevel.theory_of state)
                                        handle e => "" 

                           val sigma' = 
                          (if (trgnm = thynm) 
                           then Sigmorph.cod_ext sigma (Toplevel.theory_of state) 
                           else Sigmorph.cod_ext sigma (theory trgnm)) 

                           val sigma_s = Sigmorph.sigmorph_to_string sigma'
                           val _ = Aux.info ("Signature morphism " ^ arg ^ " : " ^ 
                                     srcnm ^ " --> " ^ trgnm ^ ", where\n" ^ sigma_s)
 
                           val _ = Sigmorph.add_sigmorph (arg, sigma')
                       in
                        ()
                       end)
                    )

fun print_sigmorph_parser tokens = (OuterParse.name tokens) |>> (Toplevel.no_timing oo print_sigmorph_trans)

val _ = OuterSyntax.improper_command "print_sigmorph"
"Print the given signature morphism" OuterKeyword.diag print_sigmorph_parser;


fun print_thymorph_trans arg =
Toplevel.keep ( fn state =>
                      (let val tmorph = case (get_thymorph arg) of
                                        SOME t => t
                                      | NONE   => error ("Unknown theory morphism " ^ arg)
                           val srcnm  = Aux.name_of_thy (Sigmorph.src (sigmorph tmorph))
                           val trgnm  = Aux.name_of_thy (Sigmorph.trg (sigmorph tmorph))
                           val thynm  = Aux.name_of_thy (Toplevel.theory_of state)
                                        handle e => ""
 
                           val tmorph' = 
                          (if (trgnm = thynm) then cod_ext tmorph Symtab.empty (Toplevel.theory_of state)
                           else cod_ext tmorph Symtab.empty (theory trgnm)) 


                           val tmorph_s = thymorph_to_string tmorph'
                           val _ = Aux.info ("Theory morphism " ^ arg ^ " : " ^ 
                                     srcnm ^ " --> " ^ trgnm ^ ", where\n" ^ tmorph_s) 

                           val _ = add_thymorph (arg, tmorph')
                       in
                        ()
                       end)
                    )

fun print_thymorph_parser tokens = (OuterParse.name tokens) |>> (Toplevel.no_timing oo print_thymorph_trans)

val _ = OuterSyntax.improper_command "print_thymorph"
"Print the given theory morphism" OuterKeyword.diag print_thymorph_parser;



(* Renaming *)


fun rename_thymorph_trans args =
Toplevel.keep ( fn state => (let val (nm, nnm) = args
                                 val tmorph = case (get_thymorph nm) of
                                                SOME t => t
                                              | NONE   => error ("Unknown theory morphism " ^ nm)
                                 val _ = delete_thymorph nm
                                 val _ = add_thymorph_verb (nnm, tmorph)
                                 val _ = Aux.info ("Theory morphism " ^ nm ^ " renamed to " ^ nnm)
                             in
                              ()
                             end))

fun rename_thymorph_parser tokens = 
((OuterParse.name -- (OuterParse.$$$ "to" |-- OuterParse.name)) tokens) |>> 
 (Toplevel.no_timing oo rename_thymorph_trans)

val _ = OuterSyntax.improper_command "rename_thymorph"
"Rename the given theory morphism" OuterKeyword.diag rename_thymorph_parser;
                                



(* *** Keywords for thymorph and sigmorph !!! *** *)

val add_keywords = List.app OuterKeyword.keyword


val _ = add_keywords ["class_map", "type_map", "op_map", "axiom_map"];
val _ = add_keywords ["by_sigmorph", "by_composition", "by_thymorph"];
val _ = add_keywords ["as", "along", "to"];
val _ = add_keywords ["axiomatic", "without_syntax", "renames", "mixfix"];
val _ = add_keywords ["-->", "\<longrightarrow>", "|->", "\<mapsto>"]; 
val _ = add_keywords ["o", "\<circ>"];  





end